# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Converters and comparators for convergence curves."""

import abc
import bisect
import copy
import enum
import logging
from typing import Callable, List, Literal, Optional, Protocol, Sequence, Union

import attr
import numpy as np
from vizier import pyvizier
from vizier.pyvizier import converters
from vizier.pyvizier import multimetric
from vizier.pyvizier.multimetric import xla_pareto


@attr.s(auto_attribs=True)
class ConvergenceCurve:
  """Represents multiple convergence curves on the same task."""

  xs: np.ndarray  # [T] array. All curves share the x axis.
  ys: np.ndarray  # [N x T] array where N is the number of curves.
  ylabel: str = ''  # Optional for plotting.
  xlabel: str = ''  # Optional for plotting.
  # Indicates ys should be increasing in t or decreasing.

  @enum.unique
  class YTrend(enum.Enum):
    """Trend of ys across t."""

    UNKNOWN = 'unknown'
    INCREASING = 'increasing'
    DECREASING = 'decreasing'

  trend: YTrend = YTrend.UNKNOWN

  def __attrs_post_init__(self):
    if len(self.ys.shape) != 2:
      raise ValueError(
          'The shape of ys should be (num_curves, n_steps); but ys shape is'
          f' {self.ys.shape}'
      )
    if len(self.xs) != self.ys.shape[1]:
      raise ValueError(
          f'Shape mismatch for time dim: {len(self.xs)} vs {self.ys.shape}'
      )
    # Allow for small numerical imprecisions.
    if self.trend == ConvergenceCurve.YTrend.INCREASING:
      if not np.all(np.nan_to_num(np.diff(self.ys, axis=-1)) >= -1e-8):
        raise ValueError(f'Increasing trend not found: {self.ys}')

    if self.trend == ConvergenceCurve.YTrend.DECREASING:
      if not np.all(np.nan_to_num(np.diff(self.ys, axis=-1)) <= 1e-8):
        raise ValueError(f'Decreasing trend not found: {self.ys}')

  @property
  def num_curves(self) -> int:
    return self.ys.shape[0]

  @classmethod
  def _interpolate_curves(
      cls,
      curves: Sequence['ConvergenceCurve'],
      *,
      resolution: Optional[int] = None,
      interpolate_repeats: bool = False,
  ) -> tuple[np.ndarray, list[np.ndarray]]:
    """Interpolate curves to have the same xs using linear interpolation.

    If xs are greater than xp[-1], then default is np.nan.

    Args:
      curves:
      resolution: Number of points to interpolate to. Leave it None to use the
        maximal resolution.
      interpolate_repeats: Interpolate repeated values (xs and ys) in the curve
        via linear interpolation (except for the last repeated segment).

    Returns:
      A tuple of the interpolated xs array, and a list of the interpolated ys
      arrays.
    """
    if not curves:
      raise ValueError('Empty sequence of curves.')
    if len(set([c.trend for c in curves])) > 1:
      raise ValueError('All curves must be increasing or decreasing.')
    minx = np.min([np.min(c.xs) for c in curves])
    maxx = np.max([np.max(c.xs) for c in curves])
    resolution = resolution or np.max([np.size(c.xs) for c in curves])
    xs = np.linspace(minx, maxx, resolution)

    all_ys = []
    for curve in curves:
      curve_ys = []
      for ys in curve.ys:
        if interpolate_repeats:
          _, indices = np.unique(ys, return_index=True)
          # Sorting indices from increasing order due to sign differences.
          indices = sorted(indices)
          if len(ys) - 1 not in indices:
            indices.append(len(ys) - 1)
        else:
          # Use the whole curve.
          indices = range(len(ys))
        curve_ys.append(
            np.interp(
                xs,
                np.array([curve.xs[i] for i in indices]),
                np.array([ys[i] for i in indices]),
                right=np.nan,
            )
        )
      all_ys.append(np.stack(curve_ys, axis=0))
    return xs, all_ys

  @classmethod
  def _align_xs_combine_ys(
      cls,
      curves: Sequence['ConvergenceCurve'],
      *,
      resolution: Optional[int] = None,
      interpolate_repeats: bool = False,
  ) -> list['ConvergenceCurve']:
    """Align curves (same xs) using linear interpolation and combine all ys."""
    xs, all_ys = cls._interpolate_curves(
        curves, resolution=resolution, interpolate_repeats=interpolate_repeats
    )
    # Take all non-empty ylabels.
    ylabels = list(set([c.ylabel for c in curves if c.ylabel]))
    if not ylabels:
      ylabel = ''
    elif len(ylabels) > 1:
      logging.info(
          'Curves have different ylabels: %s. None of them is selected.',
          ylabels,
      )
      ylabel = ''
    else:
      ylabel = ylabels[0]
    return [
        cls(xs=xs, ys=np.vstack(all_ys), ylabel=ylabel, trend=curves[0].trend)
    ]

  @classmethod
  def _align_xs_keep_ys(
      cls,
      curves: Sequence['ConvergenceCurve'],
      *,
      resolution: Optional[int] = None,
      interpolate_repeats: bool = False,
  ) -> list['ConvergenceCurve']:
    """Align curves (same xs) using linear interpolation and keep ys."""
    xs, all_ys = cls._interpolate_curves(
        curves, resolution=resolution, interpolate_repeats=interpolate_repeats
    )
    return [
        cls(xs=xs, ys=ys, ylabel=curves[i].ylabel, trend=curves[0].trend)
        for i, ys in enumerate(all_ys)
    ]

  @classmethod
  def align_xs(
      cls,
      curves: Sequence['ConvergenceCurve'],
      *,
      resolution: Optional[int] = None,
      interpolate_repeats: bool = False,
      keep_curves_separate: bool = False,
  ) -> list['ConvergenceCurve']:
    """Align curves (same xs) using linear interpolation."""
    if keep_curves_separate:
      return cls._align_xs_keep_ys(
          curves, resolution=resolution, interpolate_repeats=interpolate_repeats
      )
    else:
      return cls._align_xs_combine_ys(
          curves, resolution=resolution, interpolate_repeats=interpolate_repeats
      )

  @classmethod
  def extrapolate_ys(
      cls, curve: 'ConvergenceCurve', steps: int = 0
  ) -> 'ConvergenceCurve':
    """Extrapolates the future ys using a variant of linear extrapolation.

    Args:
      curve:
      steps: Number of steps to perform extrapolation.

    Returns:
      ConvergenceCurve whose xs, ys are extrapolated. Batch size remains the
      same but xs, ys now have length T + steps.
    """
    if curve.trend not in (cls.YTrend.INCREASING, cls.YTrend.DECREASING):
      raise ValueError('Curve must be increasing or decreasing.')

    all_extra_ys = []
    for ys in curve.ys:
      # Use average slope in the last half of the curve as slope extrapolate.
      ys_later_half = ys[int(len(ys) / 2) :]
      slope = np.mean(
          [
              ys_later_half[idx] - ys_later_half[idx - 1]
              for idx in range(1, len(ys_later_half))
          ]
      )

      extra_ys = np.append(
          ys, [ys[-1] + slope * (1 + step) for step in range(steps)]
      )
      all_extra_ys.append(extra_ys)

    return cls(
        xs=np.append(
            curve.xs, [curve.xs[-1] + 1 + step for step in range(steps)]
        ),
        ys=np.stack(all_extra_ys),
        ylabel=curve.ylabel,
        trend=curve.trend,
    )


@attr.define
class StatefulCurveConverter(abc.ABC):
  """Converter that updates its state upon each curve conversion.

  This is to ensure that distributive property holds:
  converter.convert(trials1 + trials2) = converter.convert(trials1) +
  converter.convert(trials2)
  """

  @abc.abstractmethod
  def convert(self, trials: Sequence[pyvizier.Trial]) -> ConvergenceCurve:
    """Returns a ConvergenceCurve corresponding to trials and updates state."""


@attr.define
class ConvergenceCurveConverter(StatefulCurveConverter):
  """Converter for Trial sequence to ConvergenceCurve.

  Attributes:
      metric_information: Information of relevant metric.
      flip_signs_for_min: If True, flips the signs of metric values to always
        maximize. Useful when desiring all increasing curves.
      cost_fn: Cost of each Trial (to determine xs in ConvergenceCurve).
      measurements_type: ['final', 'intermediate', 'all']
      batch_size: Number of trials in each batch. In each batch, the order of
        trials is ignored and the best trial is used.
      _best_yval: Best y-value seen. Updated in convert.
      _cumulative_cost: Cumulative cost of trials. Updated in convert.
  """

  metric_information: pyvizier.MetricInformation
  flip_signs_for_min: bool = attr.field(default=False, kw_only=True)
  cost_fn: Callable[[pyvizier.Trial], float] = attr.field(
      default=lambda _: 1, kw_only=True
  )
  measurements_type: str = attr.field(default='final', kw_only=True)
  batch_size: int = attr.field(default=1, kw_only=True)

  # Private attributes for stateful converts.
  _best_yval: float = attr.field(default=np.nan, kw_only=True)
  _cumulative_cost: float = attr.field(default=0.0, kw_only=True)

  def convert(self, trials: Sequence[pyvizier.Trial]) -> ConvergenceCurve:
    """Returns ConvergenceCurve with a single curve."""
    if not trials:
      raise ValueError(f'No trials provided {trials}')

    yvals = [self._best_yval]
    xvals = [self._cumulative_cost]
    candidates = []

    for i in range(0, len(trials), self.batch_size):
      batch = trials[i : i + self.batch_size]
      for trial in batch:
        if self.measurements_type in ('final', 'all'):
          if trial.final_measurement and (
              self.metric_information.name in trial.final_measurement.metrics
          ):
            candidates.append(
                trial.final_measurement.metrics[
                    self.metric_information.name
                ].value
            )
        if self.measurements_type in ('intermediate', 'all'):
          for measurement in trial.measurements:
            if self.metric_information.name in measurement.metrics:
              candidates.append(
                  measurement.metrics[self.metric_information.name].value
              )
        xvals.append(xvals[-1] + self.cost_fn(trial))

      new_yval = self.comparator([self.comparator(candidates), yvals[-1]])
      yvals.extend([new_yval] * len(batch))
      candidates = []

    yvals = np.asarray(yvals[1:])
    self._best_yval = self.comparator(yvals)
    self._cumulative_cost = xvals[-1]
    if self.metric_information.goal == pyvizier.ObjectiveMetricGoal.MAXIMIZE:
      trend = ConvergenceCurve.YTrend.INCREASING
      flipped = False
    elif self.flip_signs_for_min:
      trend = ConvergenceCurve.YTrend.INCREASING
      flipped = True
    else:
      trend = ConvergenceCurve.YTrend.DECREASING
      flipped = False
    return ConvergenceCurve(
        xs=np.asarray(xvals[1:]),
        ys=np.asarray(yvals).reshape([1, -1]) * (-1 if flipped else 1),
        trend=trend,
        ylabel=self.metric_information.name,
    )

  @property
  def comparator(self):
    """Comparator used for creating the convergence curve."""
    return np.nanmax if (
        self.metric_information.goal
        == pyvizier.ObjectiveMetricGoal.MAXIMIZE) else np.nanmin


class HypervolumeCurveConverter(StatefulCurveConverter):
  """Converts Trials to cumulative hypervolume curve for multiobjective."""

  def __init__(
      self,
      metric_informations: Sequence[pyvizier.MetricInformation],
      *,
      reference_value: Optional[np.ndarray] = None,
      num_vectors: int = 10000,
      infer_origin_factor: float = 0.0,
  ):
    """Init.

    Args:
      metric_informations:
      reference_value: Reference point value from which hypervolume is computed,
        with shape that is broadcastable with (dim,). Note that the sign is
        flipped for minimization metrics. If None, this computes the minimum of
        each objective as the reference point.
      num_vectors: Number of vectors from which hypervolume is computed.
      infer_origin_factor: When inferring the reference point, set origin to be
        minimum value - factor * (range).
    """
    if len(metric_informations) < 2:
      raise ValueError(
          'Should not use hypervolume curve with less than'
          f'two metrics {metric_informations}'
      )
    self._metric_informations = metric_informations
    self._num_vectors = num_vectors

    def create_metric_converter(mc):
      return converters.DefaultModelOutputConverter(
          mc,
          flip_sign_for_minimization_metrics=True,
          raise_errors_for_missing_metrics=False,
      )

    self._converter = converters.DefaultTrialConverter(
        parameter_converters=[],
        metric_converters=[
            create_metric_converter(mc) for mc in metric_informations
        ],
    )
    self._origin_value = reference_value
    # TODO: Speed this up with hypervolume vector tracking.
    self._min_trial_idx = 1
    self._pareto_frontier = np.empty(shape=(0, len(metric_informations)))
    self._infer_origin_factor = infer_origin_factor

  def convert(self, trials: Sequence[pyvizier.Trial]) -> ConvergenceCurve:
    """Returns ConvergenceCurve with a curve of shape 1 x len(trials)."""
    # Returns a len(trials) x number of metrics np.ndarray.
    if not trials:
      raise ValueError(f'No trials provided {trials}')

    metrics = self._converter.to_labels_array(trials)
    if self._origin_value is None:
      # Set origin to the minimum of finite values.
      origin = np.zeros(shape=(metrics.shape[1],))
      for metric_idx, metric_is_finite in enumerate(
          np.any(np.isfinite(metrics), axis=0)
      ):
        # If all metrics are infinite, leave origin unchanged at 0.
        if metric_is_finite:
          metric_arr = metrics[:, metric_idx]
          all_finite = metric_arr[np.isfinite(metric_arr)]
          min_finite = np.min(all_finite)
          max_finite = np.max(all_finite)
          origin[metric_idx] = min_finite - self._infer_origin_factor * (
              max_finite - min_finite
          )
      self._origin_value = origin
      logging.info(
          'Inferring origin_value as %s with metrics %s',
          self._origin_value,
          metrics,
      )
    else:
      if len(self._origin_value) == 1:
        origin = np.broadcast_to(self._origin_value, (metrics.shape[1],))
      else:
        if self._origin_value.shape != (metrics.shape[1],):
          raise ValueError(
              f'Metric shapes {self._origin_value.shape} do not '
              f'match: {(metrics.shape[1],)}'
          )
        origin = self._origin_value

    # Calculate cumulative hypervolume with the Pareto frontier.
    all_metrics = np.vstack(
        [self._pareto_frontier, metrics]
    )  # shape is [num_pareto_points + num_points, feature dimension]
    front = multimetric.ParetoFrontier(
        points=all_metrics,
        origin=origin,
        num_vectors=self._num_vectors,
        cum_hypervolume_base=xla_pareto.jax_cum_hypervolume_origin,
    )
    all_hv_curve = front.hypervolume(is_cumulative=True)

    # Remove the Pareto frontier add-in and update state.
    hv_curve = all_hv_curve[len(self._pareto_frontier) :]
    xs = np.asarray(
        range(self._min_trial_idx, len(hv_curve) + self._min_trial_idx)
    )
    self._min_trial_idx += len(hv_curve)
    algo = multimetric.FastParetoOptimalAlgorithm(
        xla_pareto.JaxParetoOptimalAlgorithm()
    )
    pareto_points = algo.is_pareto_optimal(points=all_metrics)
    self._pareto_frontier = all_metrics[pareto_points]

    return ConvergenceCurve(
        xs=xs,
        ys=np.asarray(hv_curve).reshape([1, -1]),
        trend=ConvergenceCurve.YTrend.INCREASING,
        ylabel='hypervolume',
    )


@attr.define(init=True)
class MultiMetricCurveConverter(StatefulCurveConverter):
  """Converts Trials to cumulative convergence curve for all multimetric studies.

  Attributes:
      metrics_config: Metrics config of the whole study.
      converter: Wrapped curve converter.
  """

  metrics_config: pyvizier.MetricsConfig = attr.field(
      validator=attr.validators.instance_of(pyvizier.MetricsConfig)
  )
  converter: Union[ConvergenceCurveConverter, HypervolumeCurveConverter] = (
      attr.field()
  )

  @classmethod
  def from_metrics_config(
      cls, metrics_config: pyvizier.MetricsConfig, **kwargs
  ) -> 'MultiMetricCurveConverter':
    """Builds MultiMetricConverter from config.

    Args:
      metrics_config:
      **kwargs: Kwargs forwarded to Converter.

    Returns:
    """
    if metrics_config.is_single_objective:
      single_metric_info = metrics_config.of_type(
          pyvizier.MetricType.OBJECTIVE
      ).item()
      converter = ConvergenceCurveConverter(single_metric_info, **kwargs)
      return cls(metrics_config, converter)
    else:
      converter = HypervolumeCurveConverter(
          list(metrics_config.of_type(pyvizier.MetricType.OBJECTIVE)), **kwargs
      )
    return cls(metrics_config, converter)

  def convert(self, trials: Sequence[pyvizier.Trial]) -> ConvergenceCurve:
    """Returns ConvergenceCurve with a curve of shape 1 x len(trials)."""
    if not trials:
      raise ValueError(f'No trials provided {trials}')

    # Add safety understanding by setting unsafe Trials to inf/-inf.
    safety_checker = multimetric.SafetyChecker(self.metrics_config)
    warped_trials = safety_checker.warp_unsafe_trials(copy.deepcopy(trials))

    return self.converter.convert(warped_trials)


@attr.define(init=True)
class RestartingCurveConverter(StatefulCurveConverter):
  """StatefulConverter that restarts the underlying stateful converters."""

  converter_factory: Callable[[], StatefulCurveConverter] = attr.field()
  # The minimum number of Trials needed before restarts occur.
  restart_min_trials: int = attr.field(
      default=10,
      validator=[attr.validators.instance_of(int), attr.validators.ge(0)],
      kw_only=True,
  )
  # The exponential rate at which restarts occur.
  restart_rate: float = attr.field(
      default=2,
      validator=attr.validators.ge(1),
      kw_only=True,
  )
  _all_trials: List[pyvizier.Trial] = attr.field(factory=list)
  _converter: Optional[StatefulCurveConverter] = attr.field(default=None)

  def convert(self, trials: Sequence[pyvizier.Trial]) -> ConvergenceCurve:
    if self._converter is None:
      self._converter = self.converter_factory()
      if self._all_trials:
        self._converter.convert(self._all_trials)

    curve = self._converter.convert(trials)
    self._all_trials.extend(trials)
    if len(self._all_trials) < self.restart_min_trials:
      return curve

    # Reset converter when refresh rate is crossed.
    log_previous_num_trials = np.log(
        1 + len(self._all_trials) - len(trials)
    ) / np.log(self.restart_rate)
    log_num_trials = np.log(1 + len(self._all_trials)) / np.log(
        self.restart_rate
    )
    if int(log_num_trials) > int(log_previous_num_trials):
      self._converter = None

    return curve


@attr.define
class ConvergenceComparator(abc.ABC):
  """Base class for convergence curve compartors.

  Attributes:
    baseline_curve: The baseline ConvergenceCurve.
    compared_curve: The compared ConvergenceCurve.
    baseline_quantile: Quantile in [0, 1] of the batched baseline curve to use
      for efficiency comparison. The higher the quantile, the better the quality
      of the baseline batch.
    compared_quantile: Quantile in [0, 1] of the batched compared curve to use
      for efficiency comparison. The higher the quantile, the better the quality
      of the baseline batch.
    name: Name of comparator.
    steps_cutoff: The number of initial steps (i.e., trials) to exclude from the
      convergence curve. This allows any-time algorithms to disregard initial
      trial performance, for example, when the study is guaranteed to run beyond
      that number of steps.
  """

  _baseline_curve: ConvergenceCurve = attr.field()
  _compared_curve: ConvergenceCurve = attr.field()
  _baseline_quantile: float = attr.field(
      default=0.5,
      validator=[attr.validators.le(1), attr.validators.ge(0)],
      kw_only=True,
  )
  _compared_quantile: float = attr.field(
      default=0.5,
      validator=[attr.validators.le(1), attr.validators.ge(0)],
      kw_only=True,
  )
  _name: str = attr.field(
      default='score', validator=attr.validators.instance_of(str), kw_only=True
  )
  _sign: float = attr.field(init=False)
  _steps_cutoff: Optional[float] = None

  def __attrs_post_init__(self):
    """Validates the curves and determines the sign.

    Raises:
      ValueError: If baseline curve is not INCREASING or DECREASING.
      ValueError: If the trends mismatch.
      ValueError: If baseline_quantile or compared_quantile are not in [0, 1].
    """
    if self._baseline_curve.trend not in (
        ConvergenceCurve.YTrend.INCREASING,
        ConvergenceCurve.YTrend.DECREASING,
    ):
      raise ValueError(
          f'Curve trend {self._baseline_curve.trend} must be either'
          'increasing or decreasing.'
      )
    if self._baseline_curve.trend != self._compared_curve.trend:
      raise ValueError(
          f'Baseline curve trend {self._baseline_curve.trend}'
          f' must match compared curve trend {self._compared_curve.trend}'
      )
    self._sign = (
        1.0
        if (self._baseline_curve.trend == ConvergenceCurve.YTrend.INCREASING)
        else -1.0
    )

  @abc.abstractmethod
  def score(self) -> float:
    """Returns a summary score for the comparison between base and compared.

    Usually, higher positive numbers mean the compared curve is better than the
    baseline and vice versa.
    """
    pass

  @abc.abstractmethod
  def curve(self) -> ConvergenceCurve:
    """Returns a score curve for each xs."""
    pass

  @property
  def name(self) -> str:
    return self._name

  def standardize_curves(
      self,
      apply_quantiles: bool = True,
  ) -> tuple[np.ndarray, np.ndarray]:
    """Standardize convergence curves.

    1. Align xs and keep each ys.
    2. Convert curves to INCREASING.
    3. Apply quantiles and impute NaN (optional).
    4. Remove convergence curve points where xs < steps_cutoff.

    Args:
      apply_quantiles: Whether to compute quantiles on the batches.

    Returns:
      The standardize baseline and compared curves. If apply_quantiles=True, the
      shape is (num_steps,), otherwise the shape is (batch_size, num_steps).
    """
    # Align the curves while keeping each ys.
    align_baseline_curve, align_compared_curve = ConvergenceCurve.align_xs(
        [self._baseline_curve, self._compared_curve],
        interpolate_repeats=False,
        keep_curves_separate=True,
    )
    # Adjust sign to increasing.
    baseline_ys = self._sign * align_baseline_curve.ys
    compared_ys = self._sign * align_compared_curve.ys

    if apply_quantiles:
      # Apply batch quantiels (notice the dimension reduction).
      baseline_ys = np.nanquantile(baseline_ys, self._baseline_quantile, axis=0)
      compared_ys = np.nanquantile(compared_ys, self._compared_quantile, axis=0)

    # Impute NaN values as -inf. This happens due to `align_xs` assigning
    # np.nan for xs that are outside the original convergence curve.
    baseline_ys = np.nan_to_num(baseline_ys, nan=-np.inf)
    compared_ys = np.nan_to_num(compared_ys, nan=-np.inf)

    # Remove convergence curve points where xs < steps_cutoff.
    if self._steps_cutoff is not None:
      baseline_cutoff_ind = np.where(
          align_baseline_curve.xs >= self._steps_cutoff
      )[0]
      compared_cutoff_ind = np.where(
          align_compared_curve.xs >= self._steps_cutoff
      )[0]
      if np.size(baseline_cutoff_ind) == 0 or np.size(compared_cutoff_ind) == 0:
        raise ValueError(
            f'fThe given steps_cutoff {self._steps_cutoff} is too high'
        )
      else:
        baseline_ys = baseline_ys[baseline_cutoff_ind[0] :]
        compared_ys = compared_ys[compared_cutoff_ind[0] :]

    return baseline_ys, compared_ys


class ConvergenceComparatorFactory(Protocol):
  """ConvergenceComparator factory interface."""

  def __call__(
      self,
      baseline_curve: ConvergenceCurve,
      compared_curve: ConvergenceCurve,
      baseline_quantile: float = 0.5,
      compared_quantile: float = 0.5,
      **kwargs,
  ) -> ConvergenceComparator:
    ...


@attr.define
class LogEfficiencyConvergenceCurveComparator(ConvergenceComparator):
  """Comparator methods for ConvergenceCurves.

  Methods in this class generally return comparison metrics for a compared curve
  against a baseline curve. Only works for curves with INCREASING or DECREASING
  trend.

  Example usage:
    baseline_curve = ConvergenceCurve(...)
    comparator = LogEfficiencyConvergenceCurveComparator(baseline_curve)
    comparator.curve(compared_curve)
  """

  max_score: float = attr.field(
      default=1.0, validator=[attr.validators.ge(0)], kw_only=True
  )
  summary_function: Callable[[np.ndarray], float] = attr.field(
      default=np.median
  )

  def curve(self) -> ConvergenceCurve:
    """Builds the log sample efficiency curve.

    The compared curve should approximately use exp(-relative efficiency)% less
    Trials than the baseline curve. Note that a positive relative effiency
    demonstrates that the compared curve is better than the baseline. Also,
    the relative efficiency CURVES are not fully symmetric due to differences
    in drops in objective values.

    Returns:
      ConvergenceCurves with ys (batch size 1) as the relative efficiency curve.

      The xs of curve is equal to the xs of baseline curve but there
      may be indices equal to positive 'inf' which indicate that no point
      in the compared curve outperforms the baseline at that index.

    Raises:
      ValueError: If the trends do mismatch.
      ValueError: If baseline_quantile or compared_quantile are not in [0, 1].
    """

    baseline_quantile = np.nanquantile(
        self._sign * self._baseline_curve.ys, self._baseline_quantile, axis=0
    )
    # This may not be [1,2,...] due to repeats.
    baseline_index_curve = build_convergence_curve(
        baseline_quantile, baseline_quantile
    )

    other_index_curve = build_convergence_curve(
        baseline_quantile,
        np.nanquantile(
            self._sign * self._compared_curve.ys,
            self._compared_quantile,
            axis=0,
        ),
    )

    ys = np.clip(
        np.log(1 + np.asarray(baseline_index_curve))
        - np.log(1 + np.asarray(other_index_curve)),
        a_min=-self.max_score,
        a_max=self.max_score,
    )
    return ConvergenceCurve(
        xs=self._baseline_curve.xs, ys=ys.reshape(1, len(ys))
    )

  def score(self) -> float:
    """Gets a finalized log efficiency score.

    The compared curve should approximately use exp(-score)% Trials compared to
    the baseline curve. Note that a high positive score demonstrates that the
    compared curve uses less Trials and is better than the baseline.

    Returns:
      Sample efficiency score. This score is symmetric and always finite when
      baseline_quantile <= compare_quantile (recommended setting).
    """
    baseline_curve = ConvergenceCurve.align_xs(
        [self._baseline_curve], interpolate_repeats=True
    )[0]
    compared_curve = ConvergenceCurve.align_xs(
        [self._compared_curve], interpolate_repeats=True
    )[0]
    # Combined curve (as the baseline) becomes the y-values at which
    # Trial efficiency is evaluated.
    combined_curve = ConvergenceCurve.align_xs(
        [baseline_curve, compared_curve]
    )[0]
    combined_curve.ys = np.nanmedian(combined_curve.ys, axis=0, keepdims=True)

    # Look ahead for exp(max_score)*T steps, as score is in the log space.
    extend_steps = int(np.exp(self.max_score) * len(self._baseline_curve.xs))
    extended_baseline = ConvergenceCurve.extrapolate_ys(
        baseline_curve, extend_steps
    )
    extended_compared = ConvergenceCurve.extrapolate_ys(
        compared_curve, extend_steps
    )
    baseline_comparator = LogEfficiencyConvergenceCurveComparator(
        baseline_curve=combined_curve,
        compared_curve=extended_baseline,
        compared_quantile=self._baseline_quantile,
    )
    efficiency_baseline = baseline_comparator.curve()
    compared_comparator = LogEfficiencyConvergenceCurveComparator(
        baseline_curve=combined_curve,
        compared_curve=extended_compared,
        compared_quantile=self._compared_quantile,
    )
    efficiency_compared = compared_comparator.curve()

    # Clip log efficiency and return median log efficiency in last half.
    diff = np.clip(
        efficiency_compared.ys, a_min=-self.max_score, a_max=self.max_score
    ) - np.clip(
        efficiency_baseline.ys, a_min=-self.max_score, a_max=self.max_score
    )
    return self.summary_function(diff)


@attr.define
class PercentageBetterConvergenceCurveComparator(ConvergenceComparator):
  """Comparator method based on percentage better.

  PercentageBetter is the average percentage of steps that one curve is better
  than the other.

  For example, assuming a study with 100 trials, a score of 0.07 means that on
  average for each 'baseline' trial the 'compared' convergence curve has already
  reached that value 7 steps before.
  """

  def _compute_directional_score(
      self, baseline: np.ndarray, compared: np.ndarray
  ) -> float:
    """Compute the percentage better score of 'compared' vs. 'baseline'.

    Note that: sum_i sum_j 1{c_j > b_i} = sum_j sum_i {b_i < c_j}. Therefore, we
    can either iterate over 'compared' and count the number of steps that
    'baseline' is worse OR we can iterate over 'baseline' and count the number
    of steps that 'compared' is better (which is the current implementation).

    Implementation
    --------------
    1. For each index of `baseline`:
      - Finds the smallest index of 'compared' that is better.
      - Compute the percentage of `compared` steps that are better.
    2. Average the percentages across all 'baseline' indices.

    The more dominante 'compared' over 'baseline' the closer the score is to
    1.0.

    Args:
      baseline: The baseline convergence curve.
      compared: The compared convergence curve.

    Returns:
      The average number of steps that compared is better than baseline (score
      is in [0, 1]).
    """
    convergence_curve = build_convergence_curve(list(baseline), list(compared))
    pct_baseline_compared = [
        (len(compared) - i) / len(compared) if i != float('inf') else 0
        for i in convergence_curve
    ]
    return np.mean(pct_baseline_compared)

  def score(self) -> float:
    """Computes the percentage better score.

    The score has two components:
    (a) sub-score quantifying how much 'compared' domniates 'baseline'.
    (b) sub-score quantifying how much 'baseline' domniates 'compared'.

    The final score is (a) - (b).

    Returns:
      The normalized percentage better convergence score [-1.0, 1.0].

    Raises:
      ValueError: If curve trends are not INCREASING or DECREASING, or not
      equal.
    """
    baseline_ys, compared_ys = self.standardize_curves()
    baseline_compared_score = self._compute_directional_score(
        baseline_ys, compared_ys
    )
    compared_baseline_score = self._compute_directional_score(
        compared_ys, baseline_ys
    )
    return baseline_compared_score - compared_baseline_score

  def curve(self) -> ConvergenceCurve:
    raise NotImplementedError('Curve not yet implemented.')


@attr.define
class WinRateConvergenceCurveComparator(ConvergenceComparator):
  """Comparator method based on convergence curves simple win rate comparison.

  The comparator has two modes of comparing convergence curves:

  1. Pairwise - Compare all pairs of repeated convergence curves and then
    compute the mean win-rate over all the steps (i.e. trial) and pairs.

  2. Quantiles - First compute the quantiles convergence curve per step across
    the repeates, and then compute the mean win-rate over the steps.

  The score ranges within [-0.5, 0.5], such that a score of  0.5 indicates that
  the 'compared' curve is better than 'baseline' across all stpes.
  """

  comparison_mode: Literal['pairwise', 'quantiles'] = 'pairwise'

  def score(self) -> float:
    return np.nanmean(self.curve().ys)

  def curve(self) -> ConvergenceCurve:
    """Computes the curve that represents the average win rate."""
    if self.comparison_mode == 'pairwise':
      baseline_ys, compared_ys = self.standardize_curves(apply_quantiles=False)
      # Compares all pairs of compared to baseline curve.
      all_comparisons = np.apply_along_axis(
          lambda base: np.mean(compared_ys > base, axis=0)
          + 0.5 * np.mean(base == compared_ys, axis=0),
          axis=1,
          arr=baseline_ys,
      )
      curve_ys = np.mean(all_comparisons, axis=0, keepdims=True) - 0.5
    elif self.comparison_mode == 'quantiles':
      baseline_ys, compared_ys = self.standardize_curves(apply_quantiles=True)
      curve_ys = np.asarray(
          compared_ys > baseline_ys, dtype='float'
      ) + 0.5 * np.asarray(compared_ys == baseline_ys, dtype='float')
      # Note that 0.5 is the natural average, so subtracting it to make
      # positive/negative score imply better/worse comparison.
      curve_ys = curve_ys[np.newaxis, :] - 0.5
    else:
      raise ValueError(f'Unknown comparison mode: {self.comparison_mode}')

    return ConvergenceCurve(xs=self._baseline_curve.xs, ys=curve_ys)


@attr.define
class OptimalityGapWinRateComparator(ConvergenceComparator):
  """Comparator method based on win-rate of the optimality gap."""

  def score(self):
    """Computes the normalized simple regert score."""
    baseline_ys, compared_ys = self.standardize_curves()
    return float(compared_ys[-1] > baseline_ys[-1])

  def curve(self) -> ConvergenceCurve:
    """Returns a score curve for each xs."""
    raise NotImplementedError('Curve not yet implemented.')


class OptimalityGapGainComparator(ConvergenceComparator):
  """Comparator method based on optimality gap gain.

  The optimality gap gain ('compared' - 'baseline') is normalized by the
  'baseline' absolute optimality gap and then truncated.
  """

  min_value: float = -0.5
  max_value: float = 1.0
  eps: float = 0.0001

  def score(self):
    """Computes the normalized simple regert score."""
    baseline_ys, compared_ys = self.standardize_curves()
    d = (compared_ys[-1] - baseline_ys[-1]) / (abs(baseline_ys[-1]) + self.eps)
    return min(max(d, self.min_value), self.max_value)

  def curve(self) -> ConvergenceCurve:
    """Returns a score curve for each xs."""
    raise NotImplementedError('Curve not yet implemented.')


class OptimalityGapWinRateComparatorFactory(ConvergenceComparatorFactory):
  """Factory class for OptimalityGapWinRateComparator."""

  def __call__(
      self,
      baseline_curve: ConvergenceCurve,
      compared_curve: ConvergenceCurve,
      baseline_quantile: float = 0.5,
      compared_quantile: float = 0.5,
      steps_cutoff: Optional[int] = None,
  ) -> ConvergenceComparator:
    return OptimalityGapWinRateComparator(
        baseline_curve=baseline_curve,
        compared_curve=compared_curve,
        baseline_quantile=baseline_quantile,
        compared_quantile=compared_quantile,
        name='optimality_gap_win_rate',
        steps_cutoff=steps_cutoff,
    )


class OptimalityGapGainComparatorFactory(ConvergenceComparatorFactory):
  """Factory class for OptimalityGapGainComparator."""

  def __call__(
      self,
      baseline_curve: ConvergenceCurve,
      compared_curve: ConvergenceCurve,
      baseline_quantile: float = 0.5,
      compared_quantile: float = 0.5,
      steps_cutoff: Optional[int] = None,
  ) -> ConvergenceComparator:
    return OptimalityGapGainComparator(
        baseline_curve=baseline_curve,
        compared_curve=compared_curve,
        baseline_quantile=baseline_quantile,
        compared_quantile=compared_quantile,
        name='optimality_gap_gain',
        steps_cutoff=steps_cutoff,
    )


@attr.define
class WinRateConvergenceCurveComparatorFactory(ConvergenceComparatorFactory):
  """Factory class for WinRateConvergenceCurveComparatorFactory."""

  comparison_mode: Literal['pairwise', 'quantiles'] = 'pairwise'

  def __call__(
      self,
      baseline_curve: ConvergenceCurve,
      compared_curve: ConvergenceCurve,
      baseline_quantile: float = 0.5,
      compared_quantile: float = 0.5,
      steps_cutoff: Optional[int] = None,
  ) -> ConvergenceComparator:
    return WinRateConvergenceCurveComparator(
        baseline_curve=baseline_curve,
        compared_curve=compared_curve,
        baseline_quantile=baseline_quantile,
        compared_quantile=compared_quantile,
        name='convergence_curve_win_rate',
        comparison_mode=self.comparison_mode,
        steps_cutoff=steps_cutoff,
    )


class LogEfficiencyConvergenceCurveComparatorFactory(
    ConvergenceComparatorFactory
):
  """Factory class for LogEfficiencyConvergenceCurveComparator."""

  def __call__(
      self,
      baseline_curve: ConvergenceCurve,
      compared_curve: ConvergenceCurve,
      baseline_quantile: float = 0.5,
      compared_quantile: float = 0.5,
      steps_cutoff: Optional[int] = None,
  ) -> ConvergenceComparator:
    return LogEfficiencyConvergenceCurveComparator(
        baseline_curve=baseline_curve,
        compared_curve=compared_curve,
        baseline_quantile=baseline_quantile,
        compared_quantile=compared_quantile,
        name='log_eff',
        steps_cutoff=steps_cutoff,
    )


class PercentageBetterConvergenceCurveComparatorFactory(
    ConvergenceComparatorFactory
):
  """Factory class for PercentageBetterConvergenceCurveComparator."""

  def __call__(
      self,
      baseline_curve: ConvergenceCurve,
      compared_curve: ConvergenceCurve,
      baseline_quantile: float = 0.5,
      compared_quantile: float = 0.5,
      steps_cutoff: Optional[int] = None,
  ) -> ConvergenceComparator:
    return PercentageBetterConvergenceCurveComparator(
        baseline_curve=baseline_curve,
        compared_curve=compared_curve,
        baseline_quantile=baseline_quantile,
        compared_quantile=compared_quantile,
        name='pct_better',
        steps_cutoff=steps_cutoff,
    )


def build_convergence_curve(
    baseline_curve: Sequence[float], compared_curve: Sequence[float]
) -> List[float]:
  """Builds a relative convergence curve (see returns for definition).

  Finds the smallest index j for each element i in 'baseline_curve'
  such that baseline_curve[i] <= compared_curve[j]. The function uses the
  'bisect_left' function to efficiently perform binary search under the
  assumption that 'baseline_curve' and 'compared_curve' are sorted in
  non-decreasing order.

  Args:
    baseline_curve: Baseline maximization convergence curve.
    compared_curve: Compared maximization convergence curve.

  Returns:
    A list of numbers where i-th (zero-index) element is the smallest "j" such
    that baseline_curve[i] <= compared_curve[j]
  """
  convergence_curve = []
  for value in baseline_curve:
    j = bisect.bisect_left(compared_curve, value)
    convergence_curve.append(j if j != len(compared_curve) else float('inf'))
  return convergence_curve



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from vizier._src.benchmarks.experimenters.synthetic import deb
from vizier._src.benchmarks.testing import experimenter_testing
from absl.testing import absltest
from absl.testing import parameterized


class DHExperimenterTest(parameterized.TestCase):

  @parameterized.parameters(
      (deb.DHExperimenter.DH1,),
      (deb.DHExperimenter.DH2,),
      (deb.DHExperimenter.DH3,),
      (deb.DHExperimenter.DH4,),
  )
  def test_e2e_smoke(self, experimenter_cls):
    experimenter_testing.assert_evaluates_random_suggestions(
        self, experimenter_cls(num_dimensions=4)
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Implementation of SQL Datastore."""

import collections
import threading
from typing import Callable, Iterable, List, Optional

from absl import logging
import sqlalchemy as sqla
from vizier._src.service import custom_errors
from vizier._src.service import datastore
from vizier._src.service import key_value_pb2
from vizier._src.service import resources
from vizier._src.service import study_pb2
from vizier._src.service import vizier_oss_pb2
from vizier.service import pyvizier as vz

from google.longrunning import operations_pb2

NotFoundError = custom_errors.NotFoundError
AlreadyExistsError = custom_errors.AlreadyExistsError


# TODO: Consider using ORM API (when fixed) to reduce code length.
class SQLDataStore(datastore.DataStore):
  """SQL Datastore."""

  def __init__(self, engine: sqla.engine.Engine):
    self._engine = engine
    self._connection = self._engine.connect()
    self._root_metadata = sqla.MetaData()
    self._owners_table = sqla.Table(
        'owners',
        self._root_metadata,
        sqla.Column('owner_name', sqla.String, primary_key=True),
    )
    self._studies_table = sqla.Table(
        'studies',
        self._root_metadata,
        sqla.Column('study_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('serialized_study', sqla.String),
    )
    self._trials_table = sqla.Table(
        'trials',
        self._root_metadata,
        sqla.Column('trial_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('trial_id', sqla.INTEGER),
        sqla.Column('serialized_trial', sqla.String),
    )
    self._suggestion_operations_table = sqla.Table(
        'suggestion_operations',
        self._root_metadata,
        sqla.Column('operation_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('client_id', sqla.String),
        sqla.Column('operation_number', sqla.INTEGER),
        sqla.Column('serialized_op', sqla.String),
    )
    self._early_stopping_operations_table = sqla.Table(
        'early_stopping_operations',
        self._root_metadata,
        sqla.Column('operation_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('trial_id', sqla.INTEGER),
        sqla.Column('serialized_op', sqla.String),
    )
    # This lock is meant to lock `execute()` calls for database types which
    # don't support multi-threading, like SQLite.
    self._lock = threading.Lock()
    self._root_metadata.create_all(self._engine)

  def _write_or_rollback(self, write_query: sqla.sql.Executable) -> None:
    """Wraps connection.execute() to roll back on write query failure.

    Args:
      write_query: The write query to execute.

    Raises:
      sqla.exc.DatabaseError: Generic database error.
    """
    try:
      self._connection.execute(write_query)
    except sqla.exc.DatabaseError as e:
      self._connection.rollback()
      raise e

  def create_study(self, study: study_pb2.Study) -> resources.StudyResource:
    study_resource = resources.StudyResource.from_name(study.name)
    owner_name = study_resource.owner_resource.name
    owner_query = self._owners_table.insert().values(owner_name=owner_name)
    study_query = self._studies_table.insert().values(
        study_name=study.name,
        owner_id=study_resource.owner_id,
        study_id=study_resource.study_id,
        serialized_study=study.SerializeToString(),
    )

    with self._lock:
      try:
        self._write_or_rollback(owner_query)
      except sqla.exc.IntegrityError:
        logging.info('Owner with name %s currently exists.', owner_name)

      try:
        self._write_or_rollback(study_query)
      except sqla.exc.IntegrityError as e:
        raise AlreadyExistsError(
            'Study with name %s already exists.' % study.name
        ) from e
      self._connection.commit()

    return study_resource

  def load_study(self, study_name: str) -> study_pb2.Study:
    query = sqla.select(self._studies_table)
    query = query.where(self._studies_table.c.study_name == study_name)

    with self._lock:
      row = self._connection.execute(query).fetchone()
      if not row:
        raise NotFoundError('Failed to find study name: %s' % study_name)
      study = study_pb2.Study.FromString(row.serialized_study)

    return study

  def update_study(self, study: study_pb2.Study) -> resources.StudyResource:
    study_resource = resources.StudyResource.from_name(study.name)

    # Exist query
    eq = sqla.select(self._studies_table)
    eq = eq.where(self._studies_table.c.study_name == study.name)
    eq = sqla.exists(eq).select()

    # Update query
    uq = sqla.update(self._studies_table)
    uq = uq.where(self._studies_table.c.study_name == study.name)
    uq = uq.values(
        study_name=study.name,
        owner_id=study_resource.owner_id,
        study_id=study_resource.study_id,
        serialized_study=study.SerializeToString(),
    )

    with self._lock:
      if not self._connection.execute(eq).fetchone()[0]:
        raise NotFoundError('Study %s does not exist.' % study.name)
      self._write_or_rollback(uq)
      self._connection.commit()

    return study_resource

  def delete_study(self, study_name: str) -> None:
    study_resource = resources.StudyResource.from_name(study_name)

    # Exist query
    eq = sqla.select(self._studies_table)
    eq = eq.where(self._studies_table.c.study_name == study_name)
    eq = sqla.exists(eq).select()

    # Delete study query
    dsq = self._studies_table.delete()
    dsq = dsq.where(self._studies_table.c.study_name == study_name)

    # Delete trials query
    dtq = self._trials_table.delete()
    dtq = dtq.where(self._trials_table.c.owner_id == study_resource.owner_id)
    dtq = dtq.where(self._trials_table.c.study_id == study_resource.study_id)

    with self._lock:
      if not self._connection.execute(eq).fetchone()[0]:
        raise NotFoundError('Study %s does not exist.' % study_name)
      self._write_or_rollback(dsq)
      self._write_or_rollback(dtq)
      self._connection.commit()

  def list_studies(self, owner_name: str) -> List[study_pb2.Study]:
    owner_id = resources.OwnerResource.from_name(owner_name).owner_id

    # Exist query
    eq = sqla.select(self._owners_table)
    eq = eq.where(self._owners_table.c.owner_name == owner_name)
    eq = sqla.exists(eq).select()

    # List query
    lq = sqla.select(self._studies_table)
    lq = lq.where(self._studies_table.c.owner_id == owner_id)

    with self._lock:
      if not self._connection.execute(eq).fetchone()[0]:
        raise NotFoundError('Owner name %s does not exist.' % owner_name)
      result = self._connection.execute(lq).fetchall()
      studies = [
          study_pb2.Study.FromString(row.serialized_study) for row in result
      ]

    return studies

  def create_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    trial_resource = resources.TrialResource.from_name(trial.name)
    query = self._trials_table.insert().values(
        trial_name=trial.name,
        owner_id=trial_resource.owner_id,
        study_id=trial_resource.study_id,
        trial_id=trial_resource.trial_id,
        serialized_trial=trial.SerializeToString(),
    )

    with self._lock:
      try:
        self._write_or_rollback(query)
      except sqla.exc.IntegrityError as e:
        raise AlreadyExistsError(
            'Trial with name %s already exists.' % trial.name
        ) from e
      self._connection.commit()

    return trial_resource

  def get_trial(self, trial_name: str) -> study_pb2.Trial:
    query = sqla.select(self._trials_table)
    query = query.where(self._trials_table.c.trial_name == trial_name)

    with self._lock:
      result = self._connection.execute(query)
      row = result.fetchone()

      if not row:
        raise NotFoundError('Failed to find trial name: %s' % trial_name)
      trial = study_pb2.Trial.FromString(row.serialized_trial)

    return trial

  def update_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    trial_resource = resources.TrialResource.from_name(trial.name)

    # Exist query
    eq = sqla.select(self._trials_table)
    eq = eq.where(self._trials_table.c.trial_name == trial.name)
    eq = sqla.exists(eq).select()

    # Update query
    uq = sqla.update(self._trials_table)
    uq = uq.where(self._trials_table.c.trial_name == trial.name)
    uq = uq.values(
        trial_name=trial.name,
        owner_id=trial_resource.owner_id,
        study_id=trial_resource.study_id,
        trial_id=trial_resource.trial_id,
        serialized_trial=trial.SerializeToString(),
    )

    with self._lock:
      if not self._connection.execute(eq).fetchone()[0]:
        raise NotFoundError('Trial %s does not exist.' % trial.name)
      self._write_or_rollback(uq)
      self._connection.commit()

    return trial_resource

  def list_trials(self, study_name: str) -> List[study_pb2.Trial]:
    study_resource = resources.StudyResource.from_name(study_name)

    # Exist query
    eq = sqla.select(self._studies_table)
    eq = eq.where(self._studies_table.c.study_name == study_name)
    eq = sqla.exists(eq).select()

    # List query
    lq = sqla.select(self._trials_table)
    lq = lq.where(self._trials_table.c.owner_id == study_resource.owner_id)
    lq = lq.where(self._trials_table.c.study_id == study_resource.study_id)

    with self._lock:
      if not self._connection.execute(eq).fetchone()[0]:
        raise NotFoundError('Study name %s does not exist.' % study_name)
      result = self._connection.execute(lq)
      trials = [
          study_pb2.Trial.FromString(row.serialized_trial) for row in result
      ]

    return trials

  def delete_trial(self, trial_name: str) -> None:
    # Exist query
    eq = sqla.select(self._trials_table)
    eq = eq.where(self._trials_table.c.trial_name == trial_name)
    eq = sqla.exists(eq).select()

    # Delete query
    dq = self._trials_table.delete()
    dq = dq.where(self._trials_table.c.trial_name == trial_name)

    with self._lock:
      if not self._connection.execute(eq).fetchone()[0]:
        raise NotFoundError('Trial %s does not exist.' % trial_name)
      self._write_or_rollback(dq)
      self._connection.commit()

  def max_trial_id(self, study_name: str) -> int:
    study_resource = resources.StudyResource.from_name(study_name)

    # Exist query
    eq = sqla.select(self._studies_table)
    eq = eq.where(self._studies_table.c.study_name == study_name)
    eq = sqla.exists(eq).select()

    # Trial ID query
    tq = sqla.func.max(self._trials_table.c.trial_id, type_=sqla.INT)
    tq = sqla.select(tq)
    tq = tq.where(self._trials_table.c.owner_id == study_resource.owner_id)
    tq = tq.where(self._trials_table.c.study_id == study_resource.study_id)

    with self._lock:
      if not self._connection.execute(eq).fetchone()[0]:
        raise NotFoundError('Study %s does not exist.' % study_name)
      potential_trial_id = self._connection.execute(tq).fetchone()[0]

    return potential_trial_id if potential_trial_id is not None else 0

  def create_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)
    query = self._suggestion_operations_table.insert().values(
        operation_name=operation.name,
        owner_id=resource.owner_id,
        study_id=resource.study_id,
        client_id=resource.client_id,
        operation_number=resource.operation_number,
        serialized_op=operation.SerializeToString(),
    )

    with self._lock:
      try:
        self._write_or_rollback(query)
      except sqla.exc.IntegrityError as e:
        raise AlreadyExistsError(
            'Suggest Op with name %s already exists.' % operation.name
        ) from e
      self._connection.commit()

    return resource

  def get_suggestion_operation(
      self, operation_name: str
  ) -> operations_pb2.Operation:
    q = sqla.select(self._suggestion_operations_table)
    q = q.where(
        self._suggestion_operations_table.c.operation_name == operation_name
    )

    with self._lock:
      row = self._connection.execute(q).fetchone()

      if not row:
        raise NotFoundError(
            'Failed to find suggest op name: %s' % operation_name
        )
      operation = operations_pb2.Operation.FromString(row.serialized_op)

    return operation

  def update_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)

    # Exist query
    eq = sqla.select(self._suggestion_operations_table)
    eq = eq.where(
        self._suggestion_operations_table.c.operation_name == operation.name
    )
    eq = sqla.exists(eq).select()

    # Update query
    uq = sqla.update(self._suggestion_operations_table)
    uq = uq.where(
        self._suggestion_operations_table.c.operation_name == operation.name
    )
    uq = uq.values(
        operation_name=operation.name,
        owner_id=resource.owner_id,
        study_id=resource.study_id,
        client_id=resource.client_id,
        operation_number=resource.operation_number,
        serialized_op=operation.SerializeToString(),
    )

    with self._lock:
      if not self._connection.execute(eq).fetchone()[0]:
        raise NotFoundError('Suggest op %s does not exist.' % operation.name)
      self._write_or_rollback(uq)
      self._connection.commit()

    return resource

  def list_suggestion_operations(
      self,
      study_name: str,
      client_id: str,
      filter_fn: Optional[Callable[[operations_pb2.Operation], bool]] = None,
  ) -> List[operations_pb2.Operation]:
    study_resource = resources.StudyResource.from_name(study_name)
    q = sqla.select(self._suggestion_operations_table)
    q = q.where(
        self._suggestion_operations_table.c.owner_id == study_resource.owner_id
    )
    q = q.where(
        self._suggestion_operations_table.c.study_id == study_resource.study_id
    )
    q = q.where(self._suggestion_operations_table.c.client_id == client_id)

    eq = sqla.exists(q).select()
    with self._lock:
      if not self._connection.execute(eq).fetchone()[0]:
        raise NotFoundError(
            'Could not find (study_name, client_id):',
            (study_resource.name, client_id),
        )
      result = self._connection.execute(q)
      all_ops = [
          operations_pb2.Operation.FromString(row.serialized_op)
          for row in result
      ]

    if filter_fn is None:
      return all_ops
    return [op for op in all_ops if filter_fn(op)]

  def max_suggestion_operation_number(
      self, study_name: str, client_id: str
  ) -> int:
    resource = resources.StudyResource.from_name(study_name)

    # Exist query
    eq = sqla.select(self._suggestion_operations_table)
    eq = eq.where(
        self._suggestion_operations_table.c.owner_id == resource.owner_id
    )
    eq = eq.where(
        self._suggestion_operations_table.c.study_id == resource.study_id
    )
    eq = eq.where(self._suggestion_operations_table.c.client_id == client_id)
    eq = sqla.exists(eq).select()

    # Max query
    mq = sqla.func.max(
        self._suggestion_operations_table.c.operation_number,
        type_=sqla.INT,
    )
    mq = sqla.select(mq)
    mq = mq.where(
        self._suggestion_operations_table.c.owner_id == resource.owner_id
    )
    mq = mq.where(
        self._suggestion_operations_table.c.study_id == resource.study_id
    )
    mq = mq.where(self._suggestion_operations_table.c.client_id == client_id)

    with self._lock:
      if not self._connection.execute(eq).fetchone()[0]:
        raise NotFoundError(
            'Could not find (study_name, client_id):', (study_name, client_id)
        )
      max_op_number = self._connection.execute(mq).fetchone()[0]

    return max_op_number

  def create_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name
    )
    query = self._early_stopping_operations_table.insert().values(
        operation_name=operation.name,
        owner_id=resource.owner_id,
        study_id=resource.study_id,
        trial_id=resource.trial_id,
        serialized_op=operation.SerializeToString(),
    )

    with self._lock:
      try:
        self._write_or_rollback(query)
      except sqla.exc.IntegrityError as e:
        raise AlreadyExistsError(
            'Early stopping op with name %s already exists.' % operation.name
        ) from e
      self._connection.commit()

    return resource

  def get_early_stopping_operation(
      self, operation_name: str
  ) -> vizier_oss_pb2.EarlyStoppingOperation:
    q = sqla.select(self._early_stopping_operations_table)
    q = q.where(
        self._early_stopping_operations_table.c.operation_name == operation_name
    )

    with self._lock:
      result = self._connection.execute(q)

      row = result.fetchone()
      if not row:
        raise NotFoundError(
            'Failed to find early stopping op name: %s' % operation_name
        )
      operation = vizier_oss_pb2.EarlyStoppingOperation.FromString(
          row.serialized_op
      )

    return operation

  def update_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name
    )

    # Exist query
    eq = sqla.select(self._early_stopping_operations_table)
    eq = eq.where(
        self._early_stopping_operations_table.c.operation_name == operation.name
    )
    eq = sqla.exists(eq).select()

    # Update query
    uq = sqla.update(self._early_stopping_operations_table)
    uq = uq.where(
        self._early_stopping_operations_table.c.operation_name == operation.name
    )
    uq = uq.values(
        operation_name=operation.name,
        owner_id=resource.owner_id,
        study_id=resource.study_id,
        trial_id=resource.trial_id,
        serialized_op=operation.SerializeToString(),
    )

    with self._lock:
      if not self._connection.execute(eq).fetchone()[0]:
        raise NotFoundError(
            'Early stopping op %s does not exist.' % operation.name
        )
      self._connection.execute(uq)
      self._connection.commit()

    return resource

  def update_metadata(
      self,
      study_name: str,
      study_metadata: Iterable[key_value_pb2.KeyValue],
      trial_metadata: Iterable[datastore.UnitMetadataUpdate],
  ) -> None:
    """Store the supplied metadata into the SQL database."""
    s_resource = resources.StudyResource.from_name(study_name)
    logging.debug('database.update_metadata s_resource= %s', s_resource)
    # Obtain original study.
    sq = sqla.select(self._studies_table)
    sq = sq.where(self._studies_table.c.study_name == study_name)

    with self._lock:
      row = self._connection.execute(sq).fetchone()
      if not row:
        raise NotFoundError('No such study:', s_resource.name)
      original_study = study_pb2.Study.FromString(row.serialized_study)

      # Store Study-related metadata into the database.
      vz.metadata_util.merge_study_metadata(
          original_study.study_spec, study_metadata
      )

      usq = sqla.update(self._studies_table)
      usq = usq.where(self._studies_table.c.study_name == study_name)
      usq = usq.values(serialized_study=original_study.SerializeToString())
      self._write_or_rollback(usq)

      # Split the trial-related metadata by Trial.
      split_metadata = collections.defaultdict(list)
      for md in trial_metadata:
        split_metadata[md.trial_id].append(md)

      # Now, we update one Trial at a time:
      for trial_id, md_list in split_metadata.items():
        t_resource = s_resource.trial_resource(trial_id)
        trial_name = t_resource.name

        # Obtain original trial.
        otq = sqla.select(self._trials_table)
        otq = otq.where(self._trials_table.c.trial_name == trial_name)
        row = self._connection.execute(otq).fetchone()
        if not row:
          self._connection.rollback()
          raise NotFoundError('No such trial:', trial_name)
        original_trial = study_pb2.Trial.FromString(row.serialized_trial)

        # Update Trial.
        vz.metadata_util.merge_trial_metadata(original_trial, md_list)
        utq = sqla.update(self._trials_table)
        utq = utq.where(self._trials_table.c.trial_name == trial_name)
        utq = utq.values(serialized_trial=original_trial.SerializeToString())
        self._write_or_rollback(utq)

      # Commit ALL changes if everything went well.
      self._connection.commit()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for `train_gp.py`."""

from typing import Callable

import jax
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import quasi_random
from vizier._src.algorithms.designers.gp import acquisitions
from vizier._src.algorithms.designers.gp import gp_models
from vizier._src.jax import types
from vizier._src.jax.models import multitask_tuned_gp_models
from vizier.jax import optimizers
from vizier.pyvizier import converters

from absl.testing import absltest
from absl.testing import parameterized

mt_type = multitask_tuned_gp_models.MultiTaskType


def _setup_lambda_search(
    f: Callable[[float], float],
    num_train: int = 100,
    num_test: int = 100,
    linear_coef: float = 0.0,
    ensemble_size: int = 1,
) -> tuple[gp_models.GPTrainingSpec, types.ModelData, types.ModelData]:
  """Sets up training state for a GP and outputs an test set for `f`.

  Args:
    f: 1D objective to be optimized, i.e. f(x), where x is a scalar in [-5., 5.)
    num_train: Number of training samples to generate.
    num_test: Number of testing samples to generate.
    linear_coef: If set, uses a linear kernel with coef `linear_coef` for the GP
    ensemble_size: Ensembles together `ensemble_size` GPs.

  Returns:
  A GP training spec.
  A generated train set.
  A generated test set.
  """
  assert num_train > 0 and num_test > 0, (
      f'Must provide a positive number of trials. Got {num_train} training and'
      f' {num_test} testing.'
  )

  search_space = vz.SearchSpace()
  search_space.root.add_float_param('x0', -5.0, 5.0)
  problem = vz.ProblemStatement(
      search_space=search_space,
      metric_information=vz.MetricsConfig(
          metrics=[
              vz.MetricInformation('obj', goal=vz.ObjectiveMetricGoal.MAXIMIZE),
          ]
      ),
  )

  converter = converters.TrialToModelInputConverter.from_problem(problem)
  quasi_random_designer = quasi_random.QuasiRandomDesigner(
      problem.search_space, seed=1
  )

  def create_model_data(
      num_entries: int,
  ) -> tuple[types.ModelData, list[vz.Trial]]:
    suggestions = quasi_random_designer.suggest(num_entries)
    obs_trials: list[vz.Trial] = []
    for idx, suggestion in enumerate(suggestions):
      trial = suggestion.to_trial(idx)
      x = suggestions[idx].parameters['x0'].value
      trial.complete(vz.Measurement(metrics={'obj': f(x)}))
      obs_trials.append(trial)

    model_data = converter.to_xy(obs_trials)
    return model_data, obs_trials

  train_data, _ = create_model_data(num_entries=num_train)
  train_spec = gp_models.GPTrainingSpec(
      ard_optimizer=optimizers.default_optimizer(),
      ard_rng=jax.random.PRNGKey(0),
      coroutine=gp_models.get_vizier_gp_coroutine(
          data=train_data, linear_coef=linear_coef
      ),
      ensemble_size=ensemble_size,
      ard_random_restarts=optimizers.DEFAULT_RANDOM_RESTARTS,
  )
  test_data, _ = create_model_data(num_entries=num_test)
  return train_spec, train_data, test_data


def _compute_mse(
    predictive: acquisitions.Predictive, test_data: types.ModelData
) -> float:
  """Computes the mean-squared error of `predictive` on `test_data."""

  pred_dist, _ = predictive.predict_with_aux(test_data.features)

  # We need this reshape to prevent a broadcast from (num_samples, ) -
  # (num_samples, 1) yielding (num_samples, num_samples) and breaking this
  # calculation.
  test_labels_reshaped = np.asarray(test_data.labels.unpad()).reshape(-1)

  mse = np.sum(np.square(pred_dist.mean() - test_labels_reshaped))
  return mse


class TrainedGPTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(linear_coef=0.0, ensemble_size=1),
      dict(linear_coef=0.4, ensemble_size=1),
      dict(linear_coef=0.0, ensemble_size=5),
      # dict(linear_coef=0.4, ensemble_size=5),  # This is flaky.
  )
  def test_mse_no_base(
      self, *, linear_coef: float = 0.0, ensemble_size: int = 1
  ):
    f = lambda x: -((x - 0.5) ** 2)
    spec, train_data, test_data = _setup_lambda_search(
        f,
        num_train=100,
        num_test=100,
        linear_coef=linear_coef,
        ensemble_size=ensemble_size,
    )
    gp = gp_models.train_gp(spec, train_data)
    mse = _compute_mse(gp, test_data)
    self.assertLess(mse, 2e-2)


class StackedResidualGPTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(linear_coef=0.0, ensemble_size=1),
      dict(linear_coef=0.4, ensemble_size=1),
      dict(linear_coef=0.0, ensemble_size=5),
      dict(linear_coef=0.4, ensemble_size=5),
  )
  def test_sequential_base_accuracy(
      self, *, linear_coef: float = 0.0, ensemble_size: int = 1
  ):
    """Tests that a good base with a bad top beats a bad independent model.

    Train a base on n = 100 samples, and a top GP with n = 5 samples.
    Combine these together into one predictor.

    Compare the MSE of this predictor with a test predictor trained on n = 5
    samples. The transfer learning enabled predictor should beat the test
    predictor.

    Args:
      linear_coef: The linear coefficient for the GP. Used for all trained GPs.
      ensemble_size: The number of GPs to ensemble together. Used for all
        trained GPs.
    """
    base_samples = 100
    bad_num_samples = 5
    num_test = 100
    f = lambda x: -((x - 0.5) ** 2)

    base_spec, base_train_data, _ = _setup_lambda_search(
        f,
        num_train=base_samples,
        num_test=num_test,
        linear_coef=linear_coef,
        ensemble_size=ensemble_size,
    )

    # This is purposefully bad with `bad_num_samples`
    top_spec, top_train_data, test_data = _setup_lambda_search(
        f,
        num_train=bad_num_samples,
        num_test=num_test,
        linear_coef=linear_coef,
        ensemble_size=ensemble_size,
    )

    # Combine the good base and the bad top into transfer learning GP.
    seq_base_gp = gp_models.train_gp(
        [base_spec, top_spec], [base_train_data, top_train_data]
    )

    # Create a purposefully-bad GP with `bad_num_samples` for comparison.
    test_gp_spec, test_gp_train_data, _ = _setup_lambda_search(
        f,
        num_train=bad_num_samples,
        num_test=num_test,
        linear_coef=linear_coef,
        ensemble_size=ensemble_size,
    )
    test_gp = gp_models.train_gp(test_gp_spec, test_gp_train_data)

    seq_base_mse = _compute_mse(seq_base_gp, test_data)
    test_mse = _compute_mse(test_gp, test_data)

    self.assertLess(seq_base_mse, test_mse)

  @parameterized.parameters(
      dict(linear_coef=0.0, ensemble_size=1),
      dict(linear_coef=0.4, ensemble_size=1),
      dict(linear_coef=0.0, ensemble_size=5),
      # dict(linear_coef=0.4, ensemble_size=5),  # This is flaky.
  )
  def test_multi_base(
      self, *, linear_coef: float = 0.0, ensemble_size: int = 1
  ):
    """Tests that multiple bases predict well.

    Train two good bases on n = 100 samples, and a top on n = 5 samples

    The MSE of this predictor should be similar to a GP trained on n = 100
    samples.

    Args:
      linear_coef: The linear coefficient for the GP. Used for all trained GPs.
      ensemble_size: The number of GPs to ensemble together. Used for all
        trained GPs.
    """
    large_n = 100
    small_n = 5
    num_test = 100
    f = lambda x: -((x - 0.5) ** 2)

    # Create a top GP with low training data
    top_spec, top_train_data, _ = _setup_lambda_search(
        f,
        num_train=small_n,
        num_test=num_test,
        linear_coef=linear_coef,
        ensemble_size=ensemble_size,
    )

    train_specs = []
    train_data = []

    for _ in range(2):
      base_spec, base_train_data, _ = _setup_lambda_search(
          f,
          num_train=large_n,
          num_test=num_test,
          linear_coef=linear_coef,
          ensemble_size=ensemble_size,
      )
      train_specs.append(base_spec)
      train_data.append(base_train_data)
    train_specs.append(top_spec)
    train_data.append(top_train_data)

    seq_base_gp = gp_models.train_gp(train_specs, train_data)

    # Create a good GP with sufficient training data
    test_gp_spec, test_gp_train_data, test_data = _setup_lambda_search(
        f,
        num_train=large_n,
        num_test=num_test,
        linear_coef=linear_coef,
        ensemble_size=ensemble_size,
    )
    test_gp = gp_models.train_gp(test_gp_spec, test_gp_train_data)

    seq_base_mse = _compute_mse(seq_base_gp, test_data)
    test_mse = _compute_mse(test_gp, test_data)

    self.assertAlmostEqual(seq_base_mse, test_mse, places=4)
    self.assertLess(test_mse, 2e-2)

  @parameterized.parameters(
      dict(linear_coef=0.0, ensemble_size=1),
      dict(linear_coef=0.4, ensemble_size=1),
      dict(linear_coef=0.0, ensemble_size=5),
      dict(linear_coef=0.4, ensemble_size=5),
  )
  def test_bad_base_resilience(
      self, *, linear_coef: float = 0.0, ensemble_size: int = 1
  ):
    """Tests that predictions are resilient to a bad base.

    Train a bad base on a fake objective with n = 100 samples.
    Trains a top predictor on the actual objective with n = 100 samples.
    Combines them into one predictor.

    The MSE of this predictor better than a GP trained purely on the bad
    objective.

    Args:
      linear_coef: The linear coefficient for the GP. Used for all trained GPs.
      ensemble_size: The number of GPs to ensemble together. Used for all
        trained GPs.
    """
    large_n = 100
    num_test = 100
    f = lambda x: -((x - 0.5) ** 2)
    fake_f = lambda x: x**5

    bad_base_spec, bad_base_train_data, _ = _setup_lambda_search(
        fake_f,
        num_train=large_n,
        num_test=num_test,
        linear_coef=linear_coef,
        ensemble_size=ensemble_size,
    )

    top_spec, top_train_data, test_data = _setup_lambda_search(
        f,
        num_train=large_n,
        num_test=num_test,
        linear_coef=linear_coef,
        ensemble_size=ensemble_size,
    )

    # Combine the good base and the bad top into transfer learning GP.
    seq_base_gp = gp_models.train_gp(
        [
            bad_base_spec,
            top_spec,
        ],
        [bad_base_train_data, top_train_data],
    )

    # Create a GP on the fake objective with sufficient training data
    test_gp_spec, test_gp_train_data, _ = _setup_lambda_search(
        fake_f,
        num_train=large_n,
        num_test=num_test,
        linear_coef=linear_coef,
        ensemble_size=ensemble_size,
    )
    test_gp = gp_models.train_gp(test_gp_spec, test_gp_train_data)

    seq_base_mse = _compute_mse(seq_base_gp, test_data)
    test_mse = _compute_mse(test_gp, test_data)

    self.assertLess(seq_base_mse, test_mse)

  def test_single_list_same_as_singleton(self):
    """Tests that `[state]` and `state` are treated the same."""
    large_n = 100
    num_test = 100
    f = lambda x: -((x - 0.5) ** 2)
    spec, train_data, test_data = _setup_lambda_search(
        f, num_train=large_n, num_test=num_test
    )

    list_gp = gp_models.train_gp(spec, train_data)
    singleton_gp = gp_models.train_gp([spec], [train_data])

    list_gp_mse = _compute_mse(list_gp, test_data)
    singleton_gp_mse = _compute_mse(singleton_gp, test_data)

    self.assertAlmostEqual(list_gp_mse, singleton_gp_mse)

  @parameterized.parameters(
      dict(multitask_type=mt_type.INDEPENDENT),
      dict(multitask_type=mt_type.SEPARABLE_NORMAL_TASK_KERNEL_PRIOR),
      dict(multitask_type=mt_type.SEPARABLE_LKJ_TASK_KERNEL_PRIOR),
      dict(multitask_type=mt_type.SEPARABLE_DIAG_TASK_KERNEL_PRIOR),
  )
  def test_multi_task(self, multitask_type: mt_type):
    search_space = vz.SearchSpace()
    search_space.root.add_float_param('x0', -5.0, 5.0)
    problem = vz.ProblemStatement(
        search_space=search_space,
        metric_information=vz.MetricsConfig(
            metrics=[
                vz.MetricInformation(
                    'obj1', goal=vz.ObjectiveMetricGoal.MAXIMIZE
                ),
                vz.MetricInformation(
                    'obj2', goal=vz.ObjectiveMetricGoal.MAXIMIZE
                ),
            ]
        ),
    )

    converter = converters.TrialToModelInputConverter.from_problem(problem)
    quasi_random_designer = quasi_random.QuasiRandomDesigner(
        problem.search_space, seed=1
    )
    num_entries = 100
    suggestions = quasi_random_designer.suggest(num_entries)
    obs_trials: list[vz.Trial] = []
    for idx, suggestion in enumerate(suggestions):
      trial = suggestion.to_trial(idx)
      x = suggestions[idx].parameters['x0'].value
      trial.complete(vz.Measurement(metrics={'obj1': x + 1, 'obj2': 2 * x - 1}))
      obs_trials.append(trial)

    train_entries = 60
    train_trials = obs_trials[:train_entries]
    test_trials = obs_trials[train_entries:]
    model_data = converter.to_xy(train_trials)
    train_spec = gp_models.GPTrainingSpec(
        ard_optimizer=optimizers.default_optimizer(),
        ard_rng=jax.random.PRNGKey(0),
        coroutine=gp_models.get_vizier_gp_coroutine(
            data=model_data, multitask_type=multitask_type
        ),
    )
    gp = gp_models.train_gp(train_spec, model_data)

    test_data = converter.to_xy(test_trials)
    pred_dist, _ = gp.predict_with_aux(test_data.features)
    mse = np.mean(np.square(pred_dist.mean() - test_data.labels.unpad()))
    self.assertLess(mse, 1e-2)


if __name__ == '__main__':
  # Jax disables float64 computations by default and will silently convert
  # float64s to float32s. We must explicitly enable float64.
  jax.config.update('jax_enable_x64', True)
  jax.config.update('jax_threefry_partitionable', False)
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Applies the noise function to each metric in final measurement."""

import functools
from typing import Callable, Optional, Sequence

import attr
import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter


@attr.define(auto_attribs=True)
class NoisyExperimenter(experimenter.Experimenter):
  """NoisyExperimenter applies noise to all metric(s) in final measurement.

  It stores the unnoised metric as metric_name + '_before_noise'. The noise
  function is a callable that is applied to metrics.
  """

  exptr: experimenter.Experimenter = attr.field()
  noise_fn: Callable[[float], float] = attr.field()

  @classmethod
  def from_type(
      cls,
      exptr: experimenter.Experimenter,
      noise_type: str,
      seed: Optional[int] = None,
  ) -> 'NoisyExperimenter':
    """Initializes noise_fn with noise given by noise_type."""
    dim = len(exptr.problem_statement().search_space.parameters)
    noise_fn = _create_noise_fn(
        noise_type,
        dimension=dim,
        seed=seed,
    )
    return cls(exptr, noise_fn)

  def problem_statement(self) -> pyvizier.ProblemStatement:
    return self.exptr.problem_statement()

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    self.exptr.evaluate(suggestions)
    for suggestion in suggestions:
      if suggestion.final_measurement is None:
        continue
      metric_dict_with_noise = {}
      for name, metric in suggestion.final_measurement.metrics.items():
        metric_dict_with_noise[name] = pyvizier.Metric(
            value=self.noise_fn(metric.value))
        metric_dict_with_noise[name + '_before_noise'] = metric
      suggestion.final_measurement.metrics = metric_dict_with_noise

  def __repr__(self):
    return f'NoisyExperimenter({self.noise_fn}) on {str(self.exptr)}'


def _create_noise_fn(
    noise: str,
    dimension: int,
    target_value: float = 1e-8,
    seed: Optional[int] = None,
) -> Callable[[float], float]:
  """Creates a noise function via NumPy.

  See https://bee22.com/resources/bbob%20noisy%20functions.pdf

  Args:
    noise: Noise specification
    dimension: Dimensionality of bbob function that the noise is applied to.
    target_value: The noise does not apply to values less than this.
    seed:

  Returns:
    Callable that returns the noisy version of the input.

  Raises:
    ValueError: if noise is not supported.
  """
  rng = np.random.default_rng(seed or 0)
  if noise == 'NO_NOISE':
    noise_fn = lambda v: v
  elif noise == 'MODERATE_GAUSSIAN':
    noise_fn = lambda v: v * rng.lognormal(0, 0.01)
  elif noise == 'SEVERE_GAUSSIAN':
    noise_fn = lambda v: v * rng.lognormal(0, 0.1)
  elif noise == 'MODERATE_UNIFORM':
    noise_fn = functools.partial(
        _uniform_noise,
        rng=rng,
        amplifying_exponent=0.01 * (0.49 + 1.0 / dimension),
        shrinking_exponent=0.01,
    )
  elif noise == 'SEVERE_UNIFORM':
    noise_fn = functools.partial(
        _uniform_noise,
        rng=rng,
        amplifying_exponent=0.1 * (0.49 + 1.0 / dimension),
        shrinking_exponent=0.1,
    )
  elif noise == 'MODERATE_SELDOM_CAUCHY':
    noise_fn = functools.partial(
        _cauchy_noise, rng=rng, noise_strength=0.01, noise_frequency=0.05
    )
  elif noise == 'SEVERE_SELDOM_CAUCHY':
    noise_fn = functools.partial(
        _cauchy_noise, rng=rng, noise_strength=0.1, noise_frequency=0.25
    )
  elif noise == 'LIGHT_ADDITIVE_GAUSSIAN':
    return functools.partial(_additive_normal_noise, rng=rng, stddev=0.01)
  elif noise == 'MODERATE_ADDITIVE_GAUSSIAN':
    return functools.partial(_additive_normal_noise, rng=rng, stddev=0.1)
  elif noise == 'SEVERE_ADDITIVE_GAUSSIAN':
    return functools.partial(_additive_normal_noise, rng=rng, stddev=1.0)
  else:
    raise ValueError('Noise was not supported: {}'.format(noise))
  return lambda v: _stabilized_noise(v, noise_fn, target_value)


def _uniform_noise(
    value: float,
    amplifying_exponent: float,
    shrinking_exponent: float,
    rng: np.random.Generator,
    epsilon: float = 1e-99,
) -> float:
  """Uniform noise model for bbob-noisy benchmark.

  The noise strength increases when value is small.

  Args:
    value: Function value to apply noise to.
    amplifying_exponent: "alpha" in the paper. The higher this number is, the
      more likely it is for the noisy value to be greater than the input value.
      0 or less means the noise never amplifies the function value.
    shrinking_exponent: "beta" in the paper. The higher this number is, the more
      likely it is for the noisy value to be less than the input value. 0 or
      less means the noise never shrinks the function value.
    rng: Rng.
    epsilon: "epsilon" in the paper. Prevents division by zero.

  Returns:
    Noisy version of value.
  """
  f1 = np.power(rng.uniform(), np.max([0.0, shrinking_exponent]))
  f2 = np.power(1e9 / (value + epsilon), amplifying_exponent * rng.uniform())
  return value * f1 * np.max([1.0, f2])


def _additive_normal_noise(
    value: float, stddev: float, rng: np.random.Generator
) -> float:
  """Additive normal noise."""
  return value + rng.normal(0.0, stddev)


def _cauchy_noise(
    value: float,
    noise_strength: float,
    noise_frequency: float,
    rng: np.random.Generator,
) -> float:
  """Cauchy noise model for bbob-noisy benchmark.

  The noise is infrequent and difficult to analyze due to large outliers.

  Args:
    value: Function value to apply noise to.
    noise_strength: "alpha" in the paper. Its absolute value determines the
      noise strength. The recommended setup as in the paper is to use a positive
      number.
    noise_frequency: "p" in the paper. Determines the probability of the noisy
      evaluation. Clipped (not explicitly but effectively) to [0, 1] range.
    rng:

  Returns:
    Noisy version of value.
  """
  noise = (rng.uniform() < noise_frequency) * rng.standard_cauchy()
  return value + noise_strength * np.max([0.0, 1000.0 + noise])


def _stabilized_noise(value: float,
                      noisy_fn: Callable[[float], float],
                      target_value: float = 1e-8) -> float:
  """Post processing of noise for bbob-noisy benchmark.

  We do not apply noise if the value is close to the global optima. This keeps
  the optimal value intact.


  Args:
    value: Function value to apply noise to.
    noisy_fn: "f_XX" in the paper. It applies noise to the input.
    target_value: If value is less than this number, then we do not apply the
      noise.

  Returns:
    value, if it is less than target_value. Otherwise, noisy version of
    value.
  """

  if value >= target_value:
    return noisy_fn(value) + 1.01 * target_value
  else:
    return value



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Temporary file for finding the optimal Yeo-Johnson transformation."""

from typing import Literal

from absl import logging
import numpy as np
from sklearn import preprocessing
from tensorflow_probability.substrates import jax as tfp

tfsb = tfp.staging.bijectors
tfb = tfp.bijectors
tfd = tfp.distributions


# TODO: Use Jax or numpy to find the optimum, and remove
# dependency on sklearn.
def optimal_transformation(
    data: np.ndarray,
    method: Literal['yeo-johnson', 'box-cox'] = 'yeo-johnson',
    *,
    standardize: bool = True) -> tfb.AutoCompositeTensorBijector:
  """Returns the power transformation with optimal parameterization.

  The optimal parameterization makes the transformed data as "normal"-esque
  as possible.

  Args:
    data: 1-D or 2-D array. If 1-D, then the bijector has batch_shape =[]. If
      2-D, then the bijector has batch shape equal to the last dimension
    method: 'yeo-johnson' or 'box-cox'. Boxcox can only be used for
      positive-only data.
    standardize: If True, returns a bijector that applies a power transform and
      then normalize so that the data maps to zero mean unit stddev normal. 1e-6
      is added to the stddev so that division by zero never happens.

  Returns:
    Bijector that maps data such that it follows a normal distribution.
    (standard normal if standardize=True).
  """
  dtype = data.dtype
  dimension = len(data.shape)
  if dimension not in {1, 2}:
    raise ValueError('Data must be 1-D or 2-D array')

  if dimension == 1:
    # PowerTransformer.fit() expects 2D array.
    data = data[:, np.newaxis]
    reduce_axis = None
  else:
    reduce_axis = 0

  if method == 'yeo-johnson':
    # For yeo-johnson, center the median to zero.
    # In the long run, we should consider identifying outliers that are very
    # far away from the optimum, and softclipping them to reasonable numbers.
    # This will help prevent them from having too much influence in deciding
    # the warping parameters.
    medians = np.median(data, axis=reduce_axis)
    shift1 = tfb.Shift(-medians)
    data = shift1(data)
  else:
    shift1 = tfb.Identity()
  lambdas = preprocessing.PowerTransformer(
      method, standardize=False).fit(data).lambdas_.astype(dtype)

  logging.info('Optimal lambda was: %s, %s', lambdas, lambdas.dtype)

  if dimension == 1:
    # Make it a scalar, so we don't end up with batch_shape = [1] in the
    # bijector.
    lambdas = lambdas.item()
  if method == 'yeo-johnson':
    # Cast the default values of `rho` and `shift` to the same dtype as `data`
    # to avoid dtype mismatch errors.
    warp = tfsb.YeoJohnson(
        lambdas, rho=np.asarray(2.0, dtype=dtype), shift=np.asarray(1.0, dtype)
    )
  elif method == 'box-cox':
    # Cast the default values of `rho` and `shift` to the same dtype as `data`
    # to avoid dtype mismatch errors.
    warp = tfsb.YeoJohnson(
        lambdas, rho=np.asarray(2.0, dtype), shift=np.asarray(0.0, dtype)
    )
  else:
    raise ValueError(f'Unknown method: {method}')

  if standardize:
    transformed = warp(data)  # 2-D array.
    shift2 = tfb.Shift(-np.mean(transformed, axis=reduce_axis))
    scale = tfb.Scale(1.0 / (np.std(transformed, axis=reduce_axis) + 1e-6))
    return tfb.Chain([scale, shift2, warp, shift1])
  else:
    return tfb.Chain([warp, shift1])



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""GP-Bandit using a Flax model and a TFP Gaussian Process.

A Python implementation of Google Vizier's GP-Bandit algorithm.
"""

# pylint: disable=logging-fstring-interpolation, g-long-lambda

import copy
import dataclasses
import datetime
import random
from typing import Optional, Sequence

from absl import logging
import attr
import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import quasi_random
from vizier._src.algorithms.designers.gp import acquisitions as acq_lib
from vizier._src.algorithms.designers.gp import gp_models
from vizier._src.algorithms.designers.gp import output_warpers
from vizier._src.algorithms.optimizers import eagle_strategy as es
from vizier._src.algorithms.optimizers import lbfgsb_optimizer as lo
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier._src.jax.models import multitask_tuned_gp_models
from vizier.jax import optimizers
from vizier.pyvizier import converters
from vizier.pyvizier.converters import padding
from vizier.utils import profiler

default_acquisition_optimizer_factory = vb.VectorizedOptimizerFactory(
    strategy_factory=es.VectorizedEagleStrategyFactory(
        eagle_config=es.EagleStrategyConfig()
    ),
    max_evaluations=75_000,
    suggestion_batch_size=25,
)

default_scoring_function_factory = acq_lib.bayesian_scoring_function_factory(
    lambda _: acq_lib.UCB()
)


def _experimental_override_allowed(fun):
  """No-op.

  Marks functions that can be easily overridden for experimentation.

  Args:
    fun:

  Returns:
    fun:
  """
  return fun


@attr.define(auto_attribs=False)
class VizierGPBandit(vza.Designer, vza.Predictor):
  """GP-Bandit using a Flax model.

  A minimal example of creating this designer:
  problem = vz.ProblemStatement(...)  # Configure a minimal problem statement.
  designer = VizierGPBandit(problem)

  Optionally set other attributes to change the defaults, e.g.:
  problem = vz.ProblemStatement(...)  # Configure a minimal problem statement.
  designer = VizierGPBandit(problem, use_trust_region=False)

  Attributes:
    problem: Must be a flat study with a single metric.
    acquisition_optimizer: Typically either a designer wrapped as an optimizer
      or a batched optimizer (like Eagle).
    ard_optimizer: An optimizer which should return a batch of hyperparameters
      to be ensembled.
    ard_random_restarts: The number of random initializations to run GP
      hyper-parameter optimization with.
    num_seed_trials: If greater than zero, first trial is the center of the
      search space. Afterwards, uses quasirandom until this number of trials are
      observed.
    scoring_function_factory: Callable that returns the scoring function to use.
    use_trust_region: Uses trust region to constrain initial exploration.
    rng: If not set, uses random numbers.
    metadata_ns: Metadata namespace that this designer writes to.
  """

  _problem: vz.ProblemStatement = attr.field(kw_only=False)
  _acquisition_optimizer_factory: vb.VectorizedOptimizerFactory = attr.field(
      default=default_acquisition_optimizer_factory,
      kw_only=True,
  )
  _ard_optimizer: optimizers.Optimizer[types.ParameterDict] = attr.field(
      factory=optimizers.default_optimizer,
      kw_only=True,
  )
  _ard_random_restarts: int = attr.field(
      default=optimizers.DEFAULT_RANDOM_RESTARTS, kw_only=True
  )
  _num_seed_trials: int = attr.field(default=1, kw_only=True)
  # If used, should set to 1.0 as prior uses a sum of Matern and linear but ARD
  # still tunes its amplitude. Only used for single-objective.
  _linear_coef: Optional[float] = attr.field(default=None, kw_only=True)
  _scoring_function_factory: acq_lib.ScoringFunctionFactory = attr.field(
      default=default_scoring_function_factory,
      kw_only=True,
  )
  _scoring_function_is_parallel: bool = attr.field(default=False, kw_only=True)
  # Whether to pad all inputs, and what type of schedule to use. This is to
  # ensure fewer JIT compilation passes. (Default implies no padding.)
  _padding_schedule: padding.PaddingSchedule = attr.field(
      factory=padding.PaddingSchedule, kw_only=True
  )
  _use_trust_region: bool = attr.field(default=True, kw_only=True)
  _rng: jax.Array = attr.field(
      factory=lambda: jax.random.PRNGKey(random.getrandbits(32)), kw_only=True
  )
  _metadata_ns: str = attr.field(
      default='oss_gp_bandit', kw_only=True, init=False
  )
  _ensemble_size: Optional[int] = attr.field(default=1, kw_only=True)
  _output_warper: output_warpers.OutputWarper = attr.field(
      factory=output_warpers.create_default_warper, kw_only=True
  )

  # Multi-objective parameters.
  _num_scalarizations: int = attr.field(default=1000, kw_only=True)
  _ref_scaling: float = attr.field(default=0.01, kw_only=True)
  _multitask_type: multitask_tuned_gp_models.MultiTaskType = attr.field(
      default=multitask_tuned_gp_models.MultiTaskType.INDEPENDENT,
      kw_only=True,
  )

  # ------------------------------------------------------------------
  # Internal attributes which should not be set by callers.
  # ------------------------------------------------------------------
  _trials: list[vz.Trial] = attr.field(factory=list, init=False)
  # The number of trials that have been incorporated
  # into the designer state (Cholesky decomposition, ARD).
  _incorporated_trials_count: int = attr.field(
      default=0, kw_only=True, init=False
  )
  _acquisition_optimizer: vb.VectorizedOptimizer = attr.field(init=False)

  _last_computed_gp: gp_models.GPState = attr.field(init=False)

  # The prior GP used in transfer learning. `last_computed_gp` is trained
  # on the residuals of `_prior_gp`, if one is trained.
  _prior_gp: Optional[gp_models.GPState] = attr.field(init=False, default=None)

  def __attrs_post_init__(self):
    # Extra validations
    if self._problem.search_space.is_conditional:
      raise ValueError(f'{type(self)} does not support conditional search.')
    if self._problem.search_space.num_parameters() == 0:
      raise ValueError(
          'SearchSpace should contain at least one parameter config.'
      )

    # Extra initializations.
    # Discrete parameters are continuified to account for their actual values.
    self._converter = converters.TrialToModelInputConverter.from_problem(
        self._problem,
        scale=True,
        max_discrete_indices=0,
        flip_sign_for_minimization_metrics=True,
        padding_schedule=self._padding_schedule,
    )
    self._quasi_random_sampler = quasi_random.QuasiRandomDesigner(
        self._problem.search_space,
        seed=int(jax.random.randint(self._rng, [], 0, 2**16)),
    )

    self._acquisition_optimizer = self._acquisition_optimizer_factory(
        self._converter
    )
    self._acquisition_problem = copy.deepcopy(self._problem)
    empty_data = types.ModelData(
        features=self._converter.to_features([]),
        labels=types.PaddedArray.as_padded(
            np.zeros((0, len(self._problem.metric_information)))
        ),
    )

    # Multi-objective overrides.
    m_info = self._problem.metric_information
    if not m_info.is_single_objective:

      # Create scalarization weights.
      self._rng, weights_rng = jax.random.split(self._rng)

      def acq_fn_factory(data: types.ModelData) -> acq_lib.AcquisitionFunction:
        # Scalarized UCB.
        scalarizer = acq_lib.create_hv_scalarization(
            self._num_scalarizations, data.labels, weights_rng
        )

        labels_array = data.labels.padded_array
        has_labels = labels_array.shape[0] > 0
        max_scalarized = None
        if has_labels:
          max_scalarized = jnp.max(scalarizer(labels_array), axis=-1)

        return acq_lib.ScalarizeOverAcquisitions(
            acquisition_fn=acq_lib.UCB(),
            scalarizer=scalarizer,
            reduction_fn=lambda x: jnp.mean(x, axis=0),
            max_scalarized=max_scalarized,
        )

      self._scoring_function_factory = (
          acq_lib.bayesian_scoring_function_factory(acq_fn_factory)
      )
      self._use_trust_region = False

    # Additional validations
    coroutine = gp_models.get_vizier_gp_coroutine(
        empty_data, multitask_type=self._multitask_type
    )
    params = sp.CoroutineWithData(coroutine, empty_data).setup(self._rng)
    model = sp.StochasticProcessWithCoroutine(coroutine, params)
    predictive = sp.UniformEnsemblePredictive(
        eqx.filter_jit(model.precompute_predictive)(empty_data)
    )
    scoring_fn = self._scoring_function_factory(
        empty_data, predictive, self._use_trust_region
    )
    if (
        isinstance(scoring_fn, acq_lib.MaxValueEntropySearch)
        and self._ensemble_size > 1
    ):
      raise ValueError(
          'MaxValueEntropySearch is not supported with ensemble '
          'size greater than one.'
      )

    acquisition_function = getattr(scoring_fn, 'acquisition_fn', None)
    self._acquisition_problem.metric_information = vz.MetricsConfig()
    if isinstance(acquisition_function, acq_lib.MultiAcquisitionFunction):
      for k in acquisition_function.acquisition_fns.keys():
        metric = vz.MetricInformation(k, goal=vz.ObjectiveMetricGoal.MAXIMIZE)
        self._acquisition_problem.metric_information.append(metric)
    else:
      metric = vz.MetricInformation(
          'acquisition', goal=vz.ObjectiveMetricGoal.MAXIMIZE
      )
      self._acquisition_problem.metric_information.append(metric)

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    """Update the list of completed trials."""
    del all_active
    self._trials.extend(copy.deepcopy(completed.trials))

  def set_priors(self, prior_studies: Sequence[vza.CompletedTrials]) -> None:
    """Updates the list of prior studies for transfer learning.

    Each element is treated as a new prior study, and will be stacked in order
    received - i.e. the first entry is for the first GP, the second entry is for
    the GP trained on the residuals of the first GP, etc.

    See section 3.3 of https://dl.acm.org/doi/10.1145/3097983.3098043 for more
    information, or see `gp/gp_models.py` and `gp/transfer_learning.py`

    Transfer learning is resilient to bad priors.

    Multiple calls are permitted, but unadvised. Each call will trigger
    retraining of the prior GPs - on only the state provided to `set_priors`.
    State is not incrementally updated.

    TODO: Decide on whether this method should become part of an
    interface.

    Args:
      prior_studies: A list of lists of completed trials, with one list per
        prior study. The designer will train a prior GP for each list of prior
        trials (for each `CompletedStudy` entry), in the order received.
    """
    self._rng, ard_rng = jax.random.split(self._rng)
    prior_data = [
        self._trials_to_data(prior_study.trials)
        for prior_study in prior_studies
    ]
    self._prior_gp = self._train_prior_gp(priors=prior_data, ard_rng=ard_rng)

  @property
  def _metric_info(self) -> vz.MetricInformation:
    return self._problem.metric_information.item()

  # TODO: Check the latency of `_generate_seed_trials` and look
  # into reducing it.
  @profiler.record_runtime
  def _generate_seed_trials(self, count: int) -> Sequence[vz.TrialSuggestion]:
    """Generate seed trials.

    The first seed trial is chosen as the search space center, the rest of the
    seed trials are chosen quasi-randomly.

    Arguments:
      count: The number of seed trials.

    Returns:
      The seed trials.
    """
    seed_suggestions = []
    if not self._trials:
      # TODO: Should track number of pending suggestions
      # so we don't suggest the center more than once.
      features = self._converter.to_features([])  # to extract shape.
      # NOTE: The code below assumes that a scaled value of 0.5 corresponds
      #   to the center of the feasible range.  This is true, but only by
      #   accident; ideally, we should get the center from the converters.
      continuous = self._padding_schedule.pad_features(
          0.5 * np.ones([1, features.continuous.shape[1]])
      )
      categorical = self._padding_schedule.pad_features(
          np.zeros([1, features.categorical.shape[1]], dtype=types.INT_DTYPE)
      )
      model_input = types.ModelInput(continuous, categorical)
      parameters = self._converter.to_parameters(model_input)[0]
      suggestion = vz.TrialSuggestion(
          parameters, metadata=vz.Metadata({'seeded': 'center'})
      )
      seed_suggestions.append(suggestion)
    with profiler.timeit('quasi_random_sampler_seed_trials'):
      if (remaining_counts := count - len(seed_suggestions)) > 0:
        seed_suggestions.extend(
            self._quasi_random_sampler.suggest(remaining_counts)
        )
    return seed_suggestions

  @_experimental_override_allowed
  def _warp_labels(self, labels: types.Array) -> types.Array:
    """Subclasses can override this method for experiments."""
    return np.concatenate(
        [
            self._output_warper.warp(labels[:, i : i + 1])
            for i in range(labels.shape[1])
        ],
        axis=-1,
    )

  @profiler.record_runtime
  def _trials_to_data(self, trials: Sequence[vz.Trial]) -> types.ModelData:
    """Convert trials to scaled features and warped labels."""
    model_data = self._converter.to_xy(trials)
    logging.info(
        'Transforming the labels of shape %s. Features has shape: %s',
        model_data.labels.padded_array.shape,
        types.ContinuousAndCategorical(
            model_data.features.continuous.padded_array.shape,
            model_data.features.categorical.padded_array.shape,
        ),
    )

    # Warp the output.
    unpad_labels = np.asarray(model_data.labels.unpad())
    warped_labels = self._warp_labels(unpad_labels)

    labels = types.PaddedArray.from_array(
        warped_labels,
        model_data.labels.padded_array.shape,
        fill_value=model_data.labels.fill_value,
    )
    logging.info('Transformed the labels. Now has shape: %s', labels.shape)
    return types.ModelData(model_data.features, labels)

  @_experimental_override_allowed
  def _create_gp_spec(
      self, data: types.ModelData, ard_rng: jax.Array
  ) -> gp_models.GPTrainingSpec:
    """Overrideable creation of a training spec for a GP model."""
    return gp_models.GPTrainingSpec(
        ard_optimizer=self._ard_optimizer,
        ard_rng=ard_rng,
        coroutine=gp_models.get_vizier_gp_coroutine(
            data=data,
            linear_coef=self._linear_coef,
            multitask_type=self._multitask_type,
        ),
        ensemble_size=self._ensemble_size,
        ard_random_restarts=self._ard_random_restarts,
    )

  @_experimental_override_allowed
  def _train_prior_gp(
      self,
      priors: Sequence[types.ModelData],
      ard_rng: jax.Array,
  ):
    """Trains a transfer-learning-enabled GP with prior studies.

    Args:
      priors: Data for each sequential prior to train for transfer learning.
        Assumed to be in order of training, i.e. element 0 is priors[0] is the
        first GP trained, and priors[1] trains a GP on the residuals of the GP
        trained on priors[0], and so on.
      ard_rng: RNG to do ARD to optimize GP parameters.

    Returns:
      A trained pre-computed ensemble GP.
    """
    ard_rngs = jax.random.split(ard_rng, len(priors))

    # Order `specs` in training order, i.e. `specs[0]` is trained first.
    specs = [
        self._create_gp_spec(prior_data, ard_rngs[i])
        for i, prior_data in enumerate(priors)
    ]

    # `train_gp` expects `specs` and `data` in training order, which is how
    # they were prepared above.
    return gp_models.train_gp(spec=specs, data=priors)

  @profiler.record_runtime
  def _update_gp(self, data: types.ModelData) -> gp_models.GPState:
    """Compute the designer's GP and caches the result. No-op without new data.

    Args:
      data: Data to go into GP.

    Returns:
      `GPState` object containing the trained GP.

    1. Convert trials to features and labels.
    2. Trains a pre-computed ensemble GP.

    If no new trials were added since last call, no update will occur.
    """
    if len(self._trials) == self._incorporated_trials_count:
      # If there's no change in the number of completed trials, don't update
      # state. The assumption is that trials can't be removed.
      return self._last_computed_gp
    self._incorporated_trials_count = len(self._trials)

    self._rng, ard_rng = jax.random.split(self._rng, 2)
    spec = self._create_gp_spec(data, ard_rng)
    if self._prior_gp:
      self._last_computed_gp = gp_models.train_stacked_residual_gp(
          base_gp=self._prior_gp, spec=spec, data=data
      )
    else:
      self._last_computed_gp = gp_models.train_gp(spec=spec, data=data)

    return self._last_computed_gp

  @_experimental_override_allowed
  @profiler.record_runtime
  def _optimize_acquisition(
      self, scoring_fn: acq_lib.BayesianScoringFunction, count: int
  ) -> list[vz.Trial]:
    jax.monitoring.record_event(
        '/vizier/jax/gp_bandit/optimize_acquisition/called'
    )
    # Set up optimizer and run
    seed_features = vb.trials_to_sorted_array(self._trials, self._converter)
    acq_rng, self._rng = jax.random.split(self._rng)

    score = scoring_fn.score
    score_with_aux = scoring_fn.score_with_aux

    n_parallel = None
    if self._scoring_function_is_parallel:
      n_parallel = count
      count = 1

    acquisition_optimizer = self._acquisition_optimizer
    if not isinstance(acquisition_optimizer, lo.LBFGSBOptimizer):
      acquisition_optimizer = eqx.filter_jit(acquisition_optimizer)
    best_candidates: vb.VectorizedStrategyResults = acquisition_optimizer(
        eqx.filter_jit(score),
        prior_features=seed_features,
        count=count,
        seed=acq_rng,
        score_with_aux_fn=eqx.filter_jit(score_with_aux),
        n_parallel=n_parallel,
    )

    best_candidates = dataclasses.replace(
        best_candidates, features=best_candidates.features
    )

    # Convert best_candidates (in scaled space) into suggestions (in unscaled
    # space); also append debug information like model predictions. Output shape
    # [N, D].
    logging.info('Converting the optimization result into suggestions...')
    return vb.best_candidates_to_trials(best_candidates, self._converter)

  @profiler.record_runtime
  def suggest(self, count: int = 1) -> Sequence[vz.TrialSuggestion]:
    logging.info('Suggest called with count=%d', count)
    if count > 1 and not self._scoring_function_is_parallel:
      logging.warning(
          'GAUSSIAN_PROCESS_BANDIT currently is not optimized for batched'
          ' suggestions. Suggestions in the batch are likely to be very'
          ' similar.'
      )
    if len(self._trials) < self._num_seed_trials:
      return self._generate_seed_trials(count)

    suggest_start_time = datetime.datetime.now()
    logging.info('Updating the designer state based on trials...')
    data = self._trials_to_data(self._trials)
    gp = self._update_gp(data)

    # Define acquisition function.
    scoring_fn = self._scoring_function_factory(
        data, gp, self._use_trust_region
    )
    logging.info('Optimizing acquisition: %s', scoring_fn)
    best_trials = self._optimize_acquisition(scoring_fn, count)

    suggestions = []
    for t in best_trials:
      metadata = t.metadata.ns(self._metadata_ns).ns('devinfo')
      metadata['time_spent'] = f'{datetime.datetime.now() - suggest_start_time}'
      suggestions.append(
          vz.TrialSuggestion(parameters=t.parameters, metadata=t.metadata)
      )
    return suggestions

  @profiler.record_runtime
  def sample(
      self,
      trials: Sequence[vz.TrialSuggestion],
      rng: Optional[jax.Array] = None,
      num_samples: int = 1000,
  ) -> types.Array:
    """Returns unwarped samples from the model for any given trials.

    Arguments:
      trials: The trials where the predictions will be made.
      rng: The sampling random key.
      num_samples: The number of samples per trial.

    Returns:
      The samples in the specified trials. shape: (num_samples, num_trials)
    """
    if rng is None:
      rng = jax.random.PRNGKey(0)

    if not trials:
      return np.zeros((num_samples, 0))

    data = self._trials_to_data(self._trials)
    gp = self._update_gp(data)
    xs = self._converter.to_features(trials)
    xs = types.ModelInput(
        continuous=xs.continuous.replace_fill_value(0.0),
        categorical=xs.categorical.replace_fill_value(0),
    )
    samples = eqx.filter_jit(acq_lib.sample_from_predictive)(
        gp, xs, num_samples, key=rng
    )  # (num_samples, num_trials)
    # Scope the samples to non-padded only (there's a single padded dimension).
    samples = samples[
        :, ~(xs.continuous.is_missing[0] | xs.categorical.is_missing[0])
    ]
    # TODO: vectorize output warping.
    return np.vstack([
        self._output_warper.unwarp(samples[i][..., np.newaxis]).reshape(-1)
        for i in range(samples.shape[0])
    ])

  @profiler.record_runtime
  def predict(
      self,
      trials: Sequence[vz.TrialSuggestion],
      rng: Optional[jax.Array] = None,
      num_samples: Optional[int] = 1000,
  ) -> vza.Prediction:
    """Returns the mean and stddev for any given trials.

    The method performs sampling of the warped GP model, unwarp the samples and
    compute the empirical mean and standard deviation as an apprixmation.

    Arguments:
      trials: The trials where the predictions will be made.
      rng: The sampling random key used for approximation.
      num_samples: The number of samples used for the approximation.

    Returns:
      The predictions in the specified trials.
    """
    unwarped_samples = self.sample(trials, rng, num_samples)
    mean = np.mean(unwarped_samples, axis=0)
    stddev = np.std(unwarped_samples, axis=0)
    return vza.Prediction(mean=mean, stddev=stddev)

  @classmethod
  def from_problem(
      cls,
      problem: vz.ProblemStatement,
      seed: Optional[int] = None,
      **kwargs,
  ) -> 'VizierGPBandit':
    rng = (
        jax.random.PRNGKey(random.getrandbits(32))
        if seed is None
        else jax.random.PRNGKey(seed)
    )
    return cls(problem, rng=rng, **kwargs)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Converters for PyVizier with RayTune."""

from typing import Any, Callable, Dict, Union

from ray import tune
from ray.tune.search import sample
from vizier import pyvizier as vz
from vizier.benchmarks import experimenters


class SearchSpaceConverter:
  """Converts pyvizier.SearchSpace <-> RayTune Search Space."""

  @classmethod
  def to_dict(
      cls,
      search_space: vz.SearchSpace,
  ) -> Dict[str, Union[sample.Domain, sample.Sampler]]:
    """Converts PyVizier ProblemStatement to Ray search space."""
    param_space = {}

    for param in search_space.parameters:
      if param.type == vz.ParameterType.DOUBLE:
        lower, upper = param.bounds
        if param.scale_type == vz.ScaleType.LINEAR:
          param_space[param.name] = tune.uniform(lower, upper)
        elif param.scale_type == vz.ScaleType.LOG:
          param_space[param.name] = tune.loguniform(lower, upper)
        else:
          raise ValueError(f'DOUBLE scale {param.scale_type} not supported.')
      elif param.type == vz.ParameterType.INTEGER:
        lower, upper = param.bounds
        param_space[param.name] = tune.randint(lower, upper)
      else:
        feasible_values = param.feasible_values
        param_space[param.name] = tune.choice(feasible_values)
    return param_space

  @classmethod
  def to_vizier(
      cls,
      param_space: Dict[str, Any],
  ) -> vz.SearchSpace:
    """Converts from a Ray to a Vizier SearchSpace."""
    space = vz.SearchSpace()

    for param_name, param_config in param_space.items():
      # Find out if the parameter should be scaled.
      scale_type = None
      if isinstance(param_config, sample.Float):
        if isinstance(param_config.sampler, (sample.Grid, sample.Uniform)):
          scale_type = vz.ScaleType.LINEAR
        elif isinstance(param_config.sampler, sample.LogUniform):
          scale_type = vz.ScaleType.LOG
        elif isinstance(param_config.sampler, sample.Normal):
          raise ValueError(
              f'Normal sampler is not supported: {param_name}: {param_config}'
          )
        else:
          raise ValueError(
              f'Unknown sampler type encountered: {param_name}: {param_config}'
          )

      # Add the parameter to the search space.
      if isinstance(param_config, sample.Function):
        raise ValueError('Must use tune defined types. Functions not supported')
      elif isinstance(param_config, sample.Float):
        space.root.add_float_param(
            param_name,
            min_value=param_config.lower,
            max_value=param_config.upper,
            scale_type=scale_type,
        )
      elif isinstance(param_config, sample.Integer):
        space.root.add_int_param(
            param_name,
            min_value=param_config.lower,
            max_value=param_config.upper - 1,
        )
      elif isinstance(param_config, sample.Categorical):
        if not all([isinstance(c, str) for c in param_config.categories]):
          raise ValueError('Only string values are supported for categories')
        space.root.add_categorical_param(
            param_name, feasible_values=list(map(str, param_config.categories))
        )
      else:
        raise ValueError(
            f'Unsupported config encountered: {param_name}: {param_config}'
        )
    return space


class ExperimenterConverter:
  """Converts Experimenters to Ray Trainables."""

  @classmethod
  def to_callable(
      cls,
      experimenter: experimenters.Experimenter,
  ) -> Callable[[vz.ParameterDict], Dict[str, float]]:
    """Returns a callable that takes Parameters and returns metric values."""

    def trainable(config: vz.ParameterDict) -> Dict[str, float]:
      trial = vz.Trial(parameters=config)
      experimenter.evaluate([trial])
      if trial.final_measurement is None:
        raise ValueError(f'No final_measurement on trial{trial.id}')
      return {k: v.value for k, v in trial.final_measurement.metrics.items()}

    return trainable



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Classes for starting the Vizier Server.

Extensive tests can be found in `clients_test.py`.
"""

from concurrent import futures
import datetime
import time
from typing import Optional

import attr
import grpc
import portpicker
from vizier import pythia
from vizier._src.service import constants
from vizier._src.service import datastore
from vizier._src.service import policy_factory as service_policy_factory_lib
from vizier._src.service import pythia_service
from vizier._src.service import pythia_service_pb2_grpc
from vizier._src.service import stubs_util
from vizier._src.service import vizier_service
from vizier._src.service import vizier_service_pb2_grpc


@attr.define
class DefaultVizierServer:
  """Vizier Server which runs Pythia and Vizier Servicers in the same process.

  Both servicers have access to the others' literal class instances.

  NOTE: When using this in a test, the database_url should be in-memory
  (SQL_MEMORY_URL) since tests don't easily allow arbitrarily filepaths.
  """

  _host: str = attr.field(default='localhost')
  _database_url: Optional[str] = attr.field(
      default=constants.SQL_LOCAL_URL, kw_only=True
  )
  _policy_factory: pythia.PolicyFactory = attr.field(
      factory=service_policy_factory_lib.DefaultPolicyFactory, kw_only=True
  )
  _early_stop_recycle_period: datetime.timedelta = attr.field(
      default=datetime.timedelta(seconds=0.1), kw_only=True
  )
  _port: int = attr.field(factory=portpicker.pick_unused_port, kw_only=True)

  # Fields which should not be set by users.
  _servicer: vizier_service.VizierServicer = attr.field(init=False)
  _server: grpc.Server = attr.field(init=False)
  stub: vizier_service_pb2_grpc.VizierServiceStub = attr.field(init=False)

  @property
  def datastore(self) -> datastore.DataStore:
    return self._servicer.datastore

  @property
  def endpoint(self) -> str:
    return f'{self._host}:{self._port}'

  def __attrs_post_init__(self):
    # Setup Vizier server.
    self._servicer = vizier_service.VizierServicer(
        database_url=self._database_url,
        early_stop_recycle_period=self._early_stop_recycle_period,
    )
    self._server = grpc.server(futures.ThreadPoolExecutor(max_workers=30))
    vizier_service_pb2_grpc.add_VizierServiceServicer_to_server(
        self._servicer, self._server
    )
    self._server.add_insecure_port(self.endpoint)
    self._server.start()
    self.stub = stubs_util.create_vizier_server_stub(self.endpoint)

    # Re-set the default Pythia Service to allow custom policy factories.
    default_pythia_service = pythia_service.PythiaServicer(
        self._servicer, policy_factory=self._policy_factory
    )
    self._servicer.default_pythia_service = default_pythia_service

  def wait_for_early_stop_recycle_period(self) -> None:
    time.sleep(self._early_stop_recycle_period.total_seconds())


@attr.define
class DistributedPythiaVizierServer(DefaultVizierServer):
  """Separates Pythia from Vizier via over-the-wire distributed communication.

  This is for testing / demonstration purposes only, as in normal use-cases, the
  Pythia server should actually be created in a separate process from the Vizier
  server.
  """

  _pythia_port: int = attr.field(
      factory=portpicker.pick_unused_port, kw_only=True
  )

  # Fields which should not be set by users.
  _pythia_servicer: pythia_service.PythiaServicer = attr.field(init=False)
  _pythia_server: grpc.Server = attr.field(init=False)
  pythia_stub: pythia_service_pb2_grpc.PythiaServiceStub = attr.field(
      init=False
  )

  @property
  def pythia_endpoint(self) -> str:
    return f'{self._host}:{self._pythia_port}'

  def __attrs_post_init__(self):
    super().__attrs_post_init__()
    # Setup Pythia server.
    self._pythia_servicer = pythia_service.PythiaServicer(
        policy_factory=self._policy_factory
    )
    # `max_workers=1` is used since we can only run one Pythia thread at a time.
    self._pythia_server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    pythia_service_pb2_grpc.add_PythiaServiceServicer_to_server(
        self._pythia_servicer, self._pythia_server
    )
    self._pythia_server.add_insecure_port(self.pythia_endpoint)
    self._pythia_server.start()
    self.pythia_stub = stubs_util.create_pythia_server_stub(
        self.pythia_endpoint
    )

    # Connect Vizier and Pythia servers together.
    self._servicer.default_pythia_service = self.pythia_stub
    self._pythia_servicer.connect_to_vizier(self.endpoint)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utility functions for handling vizier metadata."""

from typing import Dict, Iterable, Literal, Optional, Tuple, Type, TypeVar, Union
from absl import logging

from vizier._src.pyvizier.shared import common
from vizier._src.pyvizier.shared import trial
from vizier._src.service import key_value_pb2
from vizier._src.service import study_pb2
from vizier._src.service import vizier_service_pb2
from google.protobuf import any_pb2
from google.protobuf.message import Message

T = TypeVar('T')


def _assign_value(
    metadatum: key_value_pb2.KeyValue, value: Union[str, any_pb2.Any, Message]
) -> None:
  """Assigns value to $metadatum."""
  if isinstance(value, str):
    metadatum.ClearField('proto')
    metadatum.value = value
  elif isinstance(value, any_pb2.Any):
    metadatum.ClearField('value')
    metadatum.proto.CopyFrom(value)
  else:
    metadatum.ClearField('value')
    metadatum.proto.Pack(value)


def assign(
    container: Union[study_pb2.StudySpec, study_pb2.Trial],
    *,
    key: str,
    ns: str,
    value: Union[str, any_pb2.Any, Message],
    mode: Literal['insert_or_assign', 'insert_or_error', 'insert'] = 'insert',
) -> Tuple[key_value_pb2.KeyValue, bool]:
  """Insert and/or assign (key, value) to container.metadata.

  Args:
    container: container.metadata must be repeated KeyValue (protobuf) field.
    key:
    ns: A namespace for the key (defaults to '', which is the user's namespace).
    value: Behavior depends on the type. `str` is copied to KeyValue.value
      `any_pb2.Any` is copied to KeyValue.proto Other types are packed to
      any_pb2.Any proto, which is then copied to KeyValue.proto.
    mode: `insert_or_assign` overrides the value if (ns, key)-pair already
      exists and `insert_or_error` raises ValueError if duplicate (ns, key)-pair
      exists. `insert` blindly inserts. This is fastest and should be used if
      the data source can be trusted.

  Returns:
    (proto, inserted) where
    proto is the protobuf that was just inserted into the $container, and
    inserted is True if the proto was newly inserted, False if it was replaced.
  """
  inserted = True

  # Find existing metadatum, unless in `insert` mode.
  existing_metadatum = None
  if mode in ('insert_or_assign', 'insert_or_error'):
    for metadatum in container.metadata:
      if metadatum.key == key and metadatum.ns == ns:
        inserted = False
        if mode == 'insert_or_error':
          raise ValueError(
              f'Duplicate (ns, key) pair: ({metadatum.ns}, {metadatum.key})'
          )
        existing_metadatum = metadatum
        break

  # If the metadatum does not exist, then add the (ns, key) pair.
  metadatum = existing_metadatum or container.metadata.add(key=key, ns=ns)
  _assign_value(metadatum, value)

  return metadatum, inserted


def get(
    container: Union[study_pb2.StudySpec, study_pb2.Trial], *, key: str, ns: str
) -> Optional[str]:
  """Returns the metadata value associated with key, or None.

  Args:
    container: A Trial of a StudySpec in protobuf form.
    key: The key of a KeyValue protobuf.
    ns: A namespace for the key (defaults to '', which is the user's namespace).
  """

  for kv in container.metadata:
    if kv.key == key and kv.ns == ns:
      if not kv.HasField('proto'):
        return kv.value
  return None


def get_proto(
    container: Union[study_pb2.StudySpec, study_pb2.Trial],
    *,
    key: str,
    ns: str,
    cls: Type[T],
) -> Optional[T]:
  """Unpacks the proto metadata into message.

  Args:
    container: (const) StudySpec or Trial to search the metadata from.
    key: (const) Lookup key of the metadata.
    ns: A namespace for the key (defaults to '', which is the user's namespace).
    cls: Pass in a proto ***class***, not a proto object.

  Returns:
    Proto message, if the value associated with the key exists and
    can be parsed into proto; None otherwise.
  """
  for kv in container.metadata:
    if kv.key == key and kv.ns == ns:
      if kv.HasField('proto'):
        message = cls()
        success = kv.proto.Unpack(message)
        return message if success else None
  return None


def make_key_value_list(
    metadata: common.Metadata,
) -> list[key_value_pb2.KeyValue]:
  """Convert $metadata to a list of KeyValue protobufs."""
  result = []
  for ns, k, v in metadata.all_items():
    item = key_value_pb2.KeyValue(key=k, ns=ns.encode())
    _assign_value(item, v)
    result.append(item)
  return result


def from_key_value_list(
    kv_s: Iterable[key_value_pb2.KeyValue],
) -> common.Metadata:
  """Converts a list of KeyValue protos into a Metadata object."""
  metadata = common.Metadata()
  for kv in kv_s:
    metadata.abs_ns(common.Namespace.decode(kv.ns))[kv.key] = (
        kv.proto if kv.HasField('proto') else kv.value
    )
  return metadata


def trial_metadata_to_update_list(
    trial_metadata: dict[int, common.Metadata]
) -> list[vizier_service_pb2.UnitMetadataUpdate]:
  """Convert a dictionary of Trial.id:Metadata to a list of UnitMetadataUpdate.

  Args:
    trial_metadata: Typically MetadataDelta.on_trials.

  Returns:
    a list of UnitMetadataUpdate objects.
  """
  result = []
  for trial_id, md in trial_metadata.items():
    for kv in make_key_value_list(md):
      # TODO: Verify this implementation.
      # Should str(trial_id) below be "resources.StudyResource.from_name(
      # study_resource_name).trial_resource(trial_id=str(trial_id)).name"?
      result.append(
          vizier_service_pb2.UnitMetadataUpdate(
              trial_id=str(trial_id), metadatum=kv
          )
      )
  return result


def study_metadata_to_update_list(
    study_metadata: common.Metadata,
) -> list[vizier_service_pb2.UnitMetadataUpdate]:
  """Convert `on_study` metadata to list of metadata update protos."""
  unit_metadata_updates = []
  for ns, k, v in study_metadata.all_items():
    unit_metadata_update = vizier_service_pb2.UnitMetadataUpdate()
    metadatum = unit_metadata_update.metadatum
    metadatum.key = k
    metadatum.ns = ns.encode()
    _assign_value(metadatum, v)
    unit_metadata_updates.append(unit_metadata_update)
  return unit_metadata_updates


def to_request_proto(
    study_resource_name: str, delta: trial.MetadataDelta
) -> vizier_service_pb2.UpdateMetadataRequest:
  """Create an UpdateMetadataRequest proto.

  Args:
    study_resource_name:
    delta:

  Returns:
  """
  request = vizier_service_pb2.UpdateMetadataRequest(name=study_resource_name)

  # Study Metadata
  request.delta.extend(study_metadata_to_update_list(delta.on_study))
  # Trial metadata
  request.delta.extend(trial_metadata_to_update_list(delta.on_trials))
  return request


def merge_study_metadata(
    study_spec: study_pb2.StudySpec,
    new_metadata: Iterable[key_value_pb2.KeyValue],
) -> None:
  """Merges $new_metadata into a Study's existing metadata."""
  metadata_dict: Dict[Tuple[str, str], key_value_pb2.KeyValue] = {}
  for kv in study_spec.metadata:
    metadata_dict[(kv.ns, kv.key)] = kv
  for kv in new_metadata:
    metadata_dict[(kv.ns, kv.key)] = kv
  study_spec.ClearField('metadata')
  study_spec.metadata.extend(
      sorted(metadata_dict.values(), key=lambda kv: (kv.ns, kv.key))
  )


def merge_trial_metadata(
    trial_proto: study_pb2.Trial,
    new_metadata: Iterable[vizier_service_pb2.UnitMetadataUpdate],
) -> None:
  """Merges $new_metadata into a Trial's existing metadata.

  Args:
    trial_proto: A representation of a Trial; this will be modified.
    new_metadata: Metadata that will add or update metadata in the Trial.
  NOTE: the metadata updates in $new_metadata should have the same ID as
    $trial_proto.
  """
  metadata_dict: Dict[Tuple[str, str], key_value_pb2.KeyValue] = {}
  for kv in trial_proto.metadata:
    metadata_dict[(kv.ns, kv.key)] = kv
  for md_update in new_metadata:
    if md_update.trial_id == trial_proto.id:
      metadata_dict[(md_update.metadatum.ns, md_update.metadatum.key)] = (
          md_update.metadatum
      )
    else:
      logging.warning(
          'Metadata associated with wrong trial: %s instead of %s',
          md_update.trial_id,
          trial_proto.id,
      )
  trial_proto.ClearField('metadata')
  trial_proto.metadata.extend(
      sorted(metadata_dict.values(), key=lambda kv: (kv.ns, kv.key))
  )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""SimpleKD experimenter.

The experimenter supports three flavors (corner, center, mixed), each of which
has a different optia location. The categorical parameter takes the same three
values, and the objective function depends on the value of the categorical
parameter which make it harder to optimize.
"""

from typing import Literal, Sequence, Union
import attrs
import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter


def _float_term(x_list: list[float]) -> float:
  """Computes the float term on a list of values.

  Args:
    x_list: Elements in the list correspond to a different dimension/Parameter.

  Returns:
    The float term accounting for all the float Parameters.
  """
  float_term = 0
  for x in x_list:
    float_term += min(
        1.6,
        (x > -0.8) * -((x - 1) ** 2)
        + (x <= -0.8) * (-(1.8**2) + 150 * (x + 0.8) ** 2),
    )
  return float_term


SimpleKDCategory = Literal['corner', 'center', 'mixed']


def _categorical_term(x: str, best_category: SimpleKDCategory) -> float:
  """Computes the categorical term."""
  if x != best_category:
    return 0
  elif x == 'corner':
    return 1
  elif x == 'center':
    return 1
  elif x == 'mixed':
    return 1.5
  raise NotImplementedError(f'Unknown categorical parameter: {x}')


_feasible_discrete_values = (1, 2, 5, 6, 8)


def _discrete_term(x_list: list[int]) -> float:
  """Computes the discrete term on a list of values."""
  discrete_term = 0
  for x in x_list:
    discrete_term += [1.2, 0.0, 0.6, 0.8, 1.0][
        _feasible_discrete_values.index(x)
    ]
  return discrete_term


def _int_term(x_list: list[int]) -> float:
  """Computes the int term on a list of values."""
  int_term = 0
  for x in x_list:
    int_term += np.power(x - 2.2, 2) / 2.0
  return int_term


@attrs.define
class SimpleKDExperimenter(experimenter.Experimenter):
  """Simple problem with arbitrary number of parameter of each type.

  SimpleKD is a carefully constructed function to trap an underexploring
  hillclimb algorithm to local optima. It is parameterized by "best_category",
  which optimization algorithm must get right in order to reach the optimum.

  Note that CATEGORICAL type has only a single parameter as there's no notion
  of hillclimb or direction for categorical parameter.

  When 'output_relative_error' is True (default case) the objective value is
  the relative error compared to the optimal objective value (which could be
  negative) and the objective goal is MINIMIZE. This option allows for using the
  experiementer results directly without needing for further processing.
  When 'output_relative_error' is False, the objecitve is the SimpleKd function
  value and the objective goal is MAXIMIZE.

  The "best_category" value refers to where the optimum occurs.
  """

  best_category: SimpleKDCategory = attrs.field()  # type: ignore
  num_float_param: int = 1
  num_discrete_param: int = 1
  num_int_param: int = 1
  output_relative_error: bool = True

  def __attrs_post_init__(self):
    if self.output_relative_error and abs(self.optimal_objective) < 1e-6:
      raise ValueError(
          f"Optimal objective is too small {self.optimal_objective}, can't"
          ' compute relative error.'
      )

  def evaluate(self, suggestions: Sequence[vz.Trial]) -> None:
    for suggestion in suggestions:
      param_values = self._get_param_values(suggestion)
      value = self._compute(param_values)
      if self.output_relative_error:
        value = abs((value - self.optimal_objective) / self.optimal_objective)
      suggestion.complete(vz.Measurement({'value': value}))

  def problem_statement(self) -> vz.ProblemStatement:
    problem = vz.ProblemStatement()
    for f in range(self.num_float_param):
      problem.search_space.root.add_float_param(f'float_{f}', -1, 1)
    for d in range(self.num_discrete_param):
      problem.search_space.root.add_discrete_param(
          f'discrete_{d}', _feasible_discrete_values
      )
    for i in range(self.num_int_param):
      problem.search_space.root.add_int_param(f'int_{i}', 1, 3)
    problem.search_space.root.add_categorical_param(
        'categorical', ('corner', 'center', 'mixed')
    )
    if self.output_relative_error:
      problem.metric_information.append(
          vz.MetricInformation(
              name='value', goal=vz.ObjectiveMetricGoal.MINIMIZE
          )
      )
    else:
      problem.metric_information.append(
          vz.MetricInformation(
              name='value', goal=vz.ObjectiveMetricGoal.MAXIMIZE
          )
      )
    return problem

  def _get_param_values(
      self, trial: vz.Trial
  ) -> dict[str, list[Union[float, int, str]]]:
    """Extract the trial parameter values by type."""

    float_list = []
    int_list = []
    discrete_list = []
    for index in range(self.num_float_param):
      float_list.append(trial.parameters[f'float_{index}'].value)
    for index in range(self.num_discrete_param):
      discrete_list.append(trial.parameters[f'discrete_{index}'].value)
    for index in range(self.num_int_param):
      int_list.append(trial.parameters[f'int_{index}'].value)

    return {
        'float': float_list,
        'int': int_list,
        'discrete': discrete_list,
        'categorical': [trial.parameters['categorical'].value],
    }

  def _compute(self, params: dict[str, list[Union[float, int, str]]]) -> float:
    """Computes the SimpleKD objective value."""
    if params['categorical'][0] == 'corner':
      return (
          _categorical_term(params['categorical'][0], self.best_category)
          + 0.8 * _float_term(params['float'])
          + _discrete_term(params['discrete'])
          + _int_term(params['int'])
      )
    elif params['categorical'][0] == 'center':
      return (
          _categorical_term(params['categorical'][0], self.best_category)
          - _float_term(params['float'])
          - _discrete_term(params['discrete'])
          - _int_term(params['int'])
      )
    elif params['categorical'][0] == 'mixed':
      return (
          _categorical_term(params['categorical'][0], self.best_category)
          + 0.8 * _float_term(params['float'])
          + _discrete_term(params['discrete'])
          - _int_term(params['int'])
      )
    else:
      raise NotImplementedError(
          f'Unknown categorical parameter: {params["categorical"]}'
      )

  @property
  def optimal_objective(self) -> float:
    """Computes the optimal objective value of a SimpleKD problem."""
    values = {}
    if self.best_category == 'corner':
      values['float'] = [-1.0 for _ in range(self.num_float_param)]
      values['int'] = [1 for _ in range(self.num_float_param)]
      values['discrete'] = [1 for _ in range(self.num_float_param)]
      values['categorical'] = ['corner']

    if self.best_category == 'center':
      values['float'] = [-0.8 for _ in range(self.num_float_param)]
      values['int'] = [2 for _ in range(self.num_float_param)]
      values['discrete'] = [2 for _ in range(self.num_float_param)]
      values['categorical'] = ['center']

    if self.best_category == 'mixed':
      values['float'] = [-1.0 for _ in range(self.num_float_param)]
      values['int'] = [2 for _ in range(self.num_float_param)]
      values['discrete'] = [1 for _ in range(self.num_float_param)]
      values['categorical'] = ['mixed']

    return self._compute(values)
